# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

This module is used to share variables, enums and functions between all modules

"""

from __future__ import annotations
from collections import defaultdict
from hashlib import md5
import re
import os
import logging
from xml.sax.saxutils import escape

from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.db import get_death_or_fallback
from gramps.gen.lib import EventType, Date
from gramps.gen.plug import BasePluginManager
from gramps.plugins.lib.libgedcom import make_gedcom_date, DATE_QUALITY
from gramps.gen.plug.report import utils
from gramps.plugins.lib.libhtml import Html

HAVE_ICU = False
HAVE_ALPHABETICINDEX = False  # separate check as this is only in ICU 4.6+
try:
    from icu import Locale

    HAVE_ICU = True
    try:
        from icu import AlphabeticIndex as icuAlphabeticIndex

        HAVE_ALPHABETICINDEX = True
    except ImportError:
        from gramps.plugins.webreport.alphabeticindex import (
            AlphabeticIndex as localAlphabeticIndex,
        )
except ImportError:
    try:
        from PyICU import Locale

        HAVE_ICU = True
        try:
            from PyICU import AlphabeticIndex as icuAlphabeticIndex

            HAVE_ALPHABETICINDEX = True
        except ImportError:
            from gramps.plugins.webreport.alphabeticindex import (
                AlphabeticIndex as localAlphabeticIndex,
            )
    except ImportError:
        pass

LOG = logging.getLogger(".NarrativeWeb")

# define clear blank line for proper styling
FULLCLEAR = Html("div", class_="fullclear", inline=True)
# define all possible web page filename extensions
_WEB_EXT = (".html", ".htm", ".shtml", ".php", ".cgi")
# used to select secured web site or not
HTTP = "http://"
HTTPS = "https://"

GOOGLE_MAPS = "https://maps.googleapis.com/maps/"
# javascript code for marker path
MARKER_PATH = """
  var marker_png = '%s';
"""

# javascript code for Google's FamilyLinks...
FAMILYLINKS = """
  var tracelife = %s;

  window.addEventListener("load", function() {
    var myLatLng = new google.maps.LatLng(%s, %s);

    var mapOptions = {
      scaleControl:    true,
      panControl:      true,
      backgroundColor: '#000000',
      zoom:            %d,
      center:          myLatLng,
      mapTypeId:       google.maps.MapTypeId.ROADMAP
    };
    var map = new google.maps.Map(document.getElementById("map_canvas"),
                                  mapOptions);

    var flightPath = new google.maps.Polyline({
      path:          tracelife,
      strokeColor:   "#FF0000",
      strokeOpacity: 1.0,
      strokeWeight:  2
    });

   flightPath.setMap(map);
  });
"""

# javascript for Google's Drop Markers...
DROPMASTERS = """
  var markers = [];
  var iterator = 0;

  var tracelife = %s;
  var map;
  var myLatLng = new google.maps.LatLng(%s, %s);

  window.addEventListener("load", function() {
    var mapOptions = {
      scaleControl: true,
      zoomControl:  true,
      zoom:         %d,
      mapTypeId:    google.maps.MapTypeId.ROADMAP,
      center:       myLatLng,
    };
    map = new google.maps.Map(document.getElementById("map_canvas"),
                              mapOptions);
  });

  function drop() {
    for (var i = 0; i < tracelife.length; i++) {
      setTimeout(function() {
        addMarker();
      }, i * 1000);
    }
  }

  function addMarker() {
    var location = tracelife[iterator];
    var myLatLng = new google.maps.LatLng(location[1], location[2]);
    var infoWindow = new google.maps.InfoWindow;

    var marker = new google.maps.Marker({
      position:  myLatLng,
      map:       map,
      draggable: true,
      title:     location[0],
      animation: google.maps.Animation.DROP
    });
    markers.push(marker);
    iterator++;
    var title = "<h2>" + location[0] + "</h2>"
    bindInfoWindow(marker, map, infoWindow, title+location[4]);
  }
  function bindInfoWindow(marker, map, infoWindow, html) {
          google.maps.event.addListener(marker, 'click', function() {
              infoWindow.setContent(html);
              infoWindow.open(map, marker);
          });
  }
"""

# javascript for Google's Markers...
MARKERS = """
  var tracelife = %s;
  var map;
  var myLatLng = new google.maps.LatLng(%s, %s);

  window.addEventListener("load", function() {
    var mapOptions = {
      scaleControl:    true,
      panControl:      true,
      backgroundColor: '#000000',
      zoom:            %d,
      center:          myLatLng,
      mapTypeId:       google.maps.MapTypeId.ROADMAP
    };
    map = new google.maps.Map(document.getElementById("map_canvas"),
                              mapOptions);
    addMarkers();
  });

  function addMarkers() {
    var bounds = new google.maps.LatLngBounds();
    var infoWindow = new google.maps.InfoWindow;

    for (var i = 0; i < tracelife.length; i++) {
      var location = tracelife[i];
      var myLatLng = new google.maps.LatLng(location[1], location[2]);

      var marker = new google.maps.Marker({
        position:  myLatLng,
        draggable: true,
        title:     location[0],
        map:       map,
        zIndex:    location[3]
      });
      var title = "<h2>" + location[0] + "</h2>"
      bindInfoWindow(marker, map, infoWindow, title+location[4]);
      bounds.extend(myLatLng);
      if ( i > 1 ) { map.fitBounds(bounds); };
    }
  }
  function bindInfoWindow(marker, map, infoWindow, html) {
          google.maps.event.addListener(marker, 'click', function() {
              infoWindow.setContent(html);
              infoWindow.open(map, marker);
          });
  }
"""

# javascript for OpenStreetMap's markers...
"""
https://openlayers.org/en/latest/examples/
"""

OSM_MARKERS = """
  window.addEventListener("load", function() {
    var map;
    var heat = '%s';
    var tracelife = %s;
    var iconStyle = new ol.style.Style({
      image: new ol.style.Icon(({
        anchor: [0.2, 48],
        anchorXUnits: 'fraction',
        anchorYUnits: 'pixels',
        opacity: 1.0,
        src: marker_png
      }))
    });
    var markerSource = new ol.source.Vector({
    });
    var centerCoord = new ol.proj.transform([%s, %s], 'EPSG:4326', 'EPSG:3857');
    var zoom = %d;
    var radius = %d;
    var blur = %d;
    for (var i = 0; i < tracelife.length; i++) {
      var loc = tracelife[i];
      var iconFeature = new ol.Feature({
       geometry: new ol.geom.Point(ol.proj.transform([loc[0], loc[1]],
                                                     'EPSG:4326', 'EPSG:3857')),
       name: loc[2],
       data: loc[3],
      });
      iconFeature.setStyle(iconStyle);
      markerSource.addFeature(iconFeature);
    }
    markerLayer = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    tooltip = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    heatmap = new ol.layer.Heatmap({
      source: markerSource,
      radius: radius,
      blur: blur,
      style: iconStyle
    });
    if (heat == "heatmap") {
      map = new ol.Map({
                   target: 'map_canvas',
                   layers: [new ol.layer.Tile({ source: new ol.source.OSM() }),
                            heatmap],
                   view: new ol.View({ center: centerCoord, zoom: zoom })
                   });
    } else {
      map = new ol.Map({
                   target: 'map_canvas',
                   layers: [new ol.layer.Tile({ source: new ol.source.OSM() }),
                            markerLayer, tooltip],
                   view: new ol.View({ center: centerCoord, zoom: zoom })
                   });
    };
"""

STAMEN_MARKERS = """
  window.addEventListener("load", function() {
    var map;
    var tracelife = %s;
    var layer = '%s';
    var iconStyle = new ol.style.Style({
      image: new ol.style.Icon(({
        anchor: [0.2, 48],
        anchorXUnits: 'fraction',
        anchorYUnits: 'pixels',
        opacity: 1.0,
        src: marker_png
      }))
    });
    var markerSource = new ol.source.Vector({
    });
    for (var i = 0; i < tracelife.length; i++) {
      var loc = tracelife[i];
      var iconFeature = new ol.Feature({
       geometry: new ol.geom.Point(ol.proj.transform([loc[0], loc[1]],
                                                     'EPSG:4326', 'EPSG:3857')),
       name: loc[2],
       data: loc[3],
      });
      iconFeature.setStyle(iconStyle);
      markerSource.addFeature(iconFeature);
    }
    var centerCoord = new ol.proj.transform([%s, %s], 'EPSG:4326', 'EPSG:3857');
    markerLayer = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    tooltip = new ol.layer.Vector({
      source: markerSource,
      style: iconStyle
    });
    map = new ol.Map({
                 target: 'map_canvas',
                 layers: [
                   new ol.layer.Tile({ source: new ol.source.Stamen({
                                         layer: layer }) }),
                          markerLayer, tooltip],
                 view: new ol.View({ center: centerCoord, zoom: %d })
                 });
"""

OPENLAYER = """
    var element = document.getElementById('popup');
    var content = document.getElementById('popup-content');
    var ptitle = document.getElementById('popup-title');
    var closer = document.getElementById('popup-closer');
    var tip = document.getElementById('tooltip');
    var tipcontent = document.getElementById('tooltip-content');
    var tltip1 = undefined;
    var tltip2 = undefined;
    var tooltip = new ol.Overlay({
      element: tip,
      positioning: 'bottom-center',
      offset: [10, 0],
    });
    map.addOverlay(tooltip);
    var popup = new ol.Overlay({
      element: element,
      positioning: 'bottom-center',
      autoPan: true,
      autoPanAnimation: { duration: 500 },
      stopEvent: true
    });
    map.addOverlay(popup);
    /**
     * Add a click handler to hide the popup.
     * @return {boolean} Don't follow the href.
     */
    closer.onclick = function() {
      popup.setPosition(undefined);
      closer.blur();
      tltip1 = undefined;
      tltip2 = undefined;
      return false;
    };
    map.on('pointermove', function(evt) {
      evt.preventDefault()
      if (tltip2 !== undefined) {
        return;
      }
      var feature = this.forEachFeatureAtPixel(evt.pixel,
                                               function(feature, layer) {
        tltip1 = feature;
        return feature;
      });
      map.getTargetElement().style.cursor = feature ? 'pointer' : '';
      if (evt.dragging) {
        tooltip.setPosition(undefined);
        return;
      }
      if (feature) {
        var coordinate = evt.coordinate;
        tipcontent.innerHTML = feature.get('name');
        tooltip.setPosition(coordinate);
      } else {
        tooltip.setPosition(undefined);
      }
    });
    map.on('singleclick', function(evt) {
      evt.preventDefault();
      if (tltip1 !== undefined) {
        tooltip.setPosition(undefined);
      }
      var feature = map.forEachFeatureAtPixel(evt.pixel,
                                              function(feature, layer) {
        tltip2 = feature;
        return feature;
      });
      if (feature) {
      var coordinate = evt.coordinate;
      var title = '<h2>' + feature.get('name') + '</h2>';
      ptitle.innerHTML = title;
      content.innerHTML = feature.get('data');
      popup.setPosition(coordinate);
      }
      });
  });
"""

# variables for alphabet_navigation()
_KEYPERSON, _KEYPLACE, _KEYEVENT, _ALPHAEVENT = 0, 1, 2, 3

COLLATE_LANG = glocale.collation

_NAME_STYLE_SHORT = 2
_NAME_STYLE_DEFAULT = 1
_NAME_STYLE_FIRST = 0
_NAME_STYLE_SPECIAL = None

PLUGMAN = BasePluginManager.get_instance()
CSS = PLUGMAN.process_plugin_data("WEBSTUFF")

# _NAME_COL = 3

_WRONGMEDIAPATH: list[list[str]] = []

_HTML_DBL_QUOTES = re.compile(r'([^"]*) " ([^"]*) " (.*)', re.VERBOSE)
_HTML_SNG_QUOTES = re.compile(r"([^']*) ' ([^']*) ' (.*)", re.VERBOSE)

# Events that are usually a family event
_EVENTMAP = set(
    [
        EventType.MARRIAGE,
        EventType.MARR_ALT,
        EventType.MARR_SETTL,
        EventType.MARR_LIC,
        EventType.MARR_CONTR,
        EventType.MARR_BANNS,
        EventType.ENGAGEMENT,
        EventType.DIVORCE,
        EventType.DIV_FILING,
    ]
)

# Names for stylesheets
_NARRATIVESCREEN = "narrative-screen.css"
_NARRATIVEPRINT = "narrative-print.css"

HOLIDAYS_ASSOC = [
    # LANG INDEX HOLIDAY_NAME
    ("be", 0, "Bulgaria"),
    ("ca", 1, "Canada"),
    ("cs", 2, "Czech Republic"),
    ("cl", 3, "Chile"),
    ("zh", 4, "China"),
    ("hr", 5, "Croatia"),
    ("en", 13, "United States of America"),  # must be here. should be en_US
    ("en_GB", 6, "England"),
    ("fi", 7, "Finland"),
    ("fr_FR", 8, "France"),
    ("de", 9, "Germany"),
    ("ja", 10, "Japan"),
    ("sk", 11, "Slovakia"),
    ("sv", 12, "Sweden"),
    ("he", 14, "Jewish Holidays"),
    ("en_NZ", 15, "New Zealand"),
    ("uk", 16, "Ukraine"),
    ("sr_RS", 17, "Serbia"),
    ("sr_RS@latin", 18, "Serbia (Latin)"),
]


def do_we_have_holidays(lang):
    """
    Associate an index for the holidays depending on the LANG
    """
    for lng, idx, dummy_name in HOLIDAYS_ASSOC:
        if lng == lang:  # i.e. en_US
            return idx
        if lng[:2] == lang:  # i.e. en_US[:2] => en
            return idx
    return None


def get_surname_from_person(dbase, person):
    """
    get the person's surname
    get the primary name
    if group as get the group_as surname
    else get the primary surname of the primary name
         and correct for [global] group_as name
    correct for surnames that are space or None
    """
    primary_name = person.get_primary_name()

    if primary_name.group_as:
        surname = primary_name.group_as
    else:
        group_map = _nd.primary_surname(primary_name)
        surname = dbase.get_name_group_mapping(group_map)

    # Treat people who have no name with those whose name is just
    # 'whitespace'
    if surname is None or surname.isspace():
        surname = ""
    return surname


def sort_people(dbase, handle_list, rlocale=glocale):
    """
    will sort the database people by surname
    @param: dbase           -- The instance of the database
    @param: handle_list     -- The list of handles of people to sort
    @param: rlocale         -- The locale related to the language used for the
                               sort
    @result:                -- A list sorted by surname, each element of which
                               consists of a tuple of (surname, list of handles)
                               where the list of handles is sorted by
                               primary surname, first name, suffix.
                               Surname uses group_as, but primary surname
                               does not.
    get the primary name
    if group as get the group_as surname
    else get the primary surname of the primary name
         and correct for [global] group_as name
    correct for surnames that are space or None
    for each surname sort handles by the surname, first name and suffix
    construct a list of surnames and list of handles
    """
    sname_sub = defaultdict(list)
    sortnames = {}

    for person_handle in handle_list:
        person = dbase.get_person_from_handle(person_handle)
        surname = get_surname_from_person(dbase, person)
        sortnames[person_handle] = _nd.sort_string(person.get_primary_name())
        sname_sub[surname].append(person_handle)

    sorted_lists = []
    temp_list = sorted(sname_sub, key=rlocale.sort_key)

    for name in temp_list:
        if isinstance(name, bytes):
            name = name.decode("utf-8")
        slist = sorted(
            ((sortnames[x], x) for x in sname_sub[name]),
            key=lambda x: rlocale.sort_key(x[0]),
        )
        entries = [x[1] for x in slist]
        sorted_lists.append((name, entries))

    return sorted_lists


def sort_places(dbase, handle_list, rlocale=glocale):
    """
    will sort the database places
    """
    pname_sub = defaultdict(list)

    for place_name in handle_list.keys():
        cname = sname = None
        if len(handle_list[place_name]) == 6:
            (hdle, pname, cname, sname, dummy_id, event) = handle_list[place_name]
        else:
            continue
        place = dbase.get_place_from_handle(hdle)
        pname = _pd.display(dbase, place, fmt=0)
        apname = _pd.display_event(dbase, event, fmt=0)

        pname_sub[pname].append(hdle)
        pname_sub[apname].append((hdle, pname, cname, sname))

    sorted_lists = []
    temp_list = sorted(pname_sub, key=rlocale.sort_key)

    for name in temp_list:
        if isinstance(name, bytes):
            name = name.decode("utf-8")
        sorted_lists.append((name, pname_sub[name][0]))

    return sorted_lists


def sort_event_types(dbase, event_types, event_handle_list, rlocale):
    """
    sort a list of event types and their associated event handles

    @param: dbase -- report database
    @param: event_types -- a dict of event types
    @param: event_handle_list -- all event handles in this database
    """
    rts = rlocale.translation.sgettext
    event_dict = dict((rts(evt_type), list()) for evt_type in event_types)

    for event_handle in event_handle_list:
        event = dbase.get_event_from_handle(event_handle)
        event_type = rts(event.get_type().xml_str())

        # add (gramps_id, date, handle) from this event
        if event_type in event_dict:
            sort_value = event.get_date_object().get_sort_value()
            event_dict[event_type].append((sort_value, event_handle))

    for tup_list in event_dict.values():
        tup_list.sort()

    # return a list of sorted tuples, one per event
    retval = [
        (event_type, event_list) for (event_type, event_list) in event_dict.items()
    ]
    retval.sort(key=lambda item: str(item[0]))

    return retval


# Modified _get_regular_surname from WebCal.py to get prefix, first name,
# and suffix
def _get_short_name(gender, name):
    """Will get suffix for all people passed through it"""

    dummy_gender = gender
    short_name = name.get_first_name()
    suffix = name.get_suffix()
    if suffix:
        # TODO for Arabic, should the next line's comma be translated?
        short_name = short_name + ", " + suffix
    return short_name


def __get_person_keyname(dbase, handle):
    """...."""

    person = dbase.get_person_from_handle(handle)
    return _nd.sort_string(person.get_primary_name())


def __get_place_keyname(dbase, handle):
    """..."""

    return utils.place_name(dbase, handle)


if HAVE_ALPHABETICINDEX:

    class AlphabeticIndex(icuAlphabeticIndex):
        """
        Call the ICU AlphabeticIndex, passing the ICU Locale
        """

        def __init__(self, rlocale):
            self.iculocale = Locale(rlocale.collation)
            super().__init__(self.iculocale)

            # set the maximum number of buckets, the undocumented default is 99
            # Latin + Greek + Cyrillic + Hebrew + Arabic + Tamil + Hiragana +
            # CJK Unified is about 206 different buckets
            self.maxLabelCount = 500  # pylint: disable=invalid-name

            # Add bucket labels for scripts other than the one for the output
            # which is being generated
            self.iculocale.addLikelySubtags()
            default_script = self.iculocale.getDisplayScript()
            used_scripts = [default_script]

            for lang_code in glocale.get_language_dict().values():
                loc = Locale(lang_code)
                loc.addLikelySubtags()
                script = loc.getDisplayScript()
                if script not in used_scripts:
                    used_scripts.append(script)
                    super().addLabels(loc)

else:
    AlphabeticIndex = localAlphabeticIndex  # type: ignore[misc, assignment]


def alphabet_navigation(
    sorted_alpha_index,
    rlocale=glocale,
    current=None,
    rtl=False,
    only=True,
    new_page=False,
    ext=None,
):
    """
    Will create the alphabet navigation bar for classes IndividualListPage,
    SurnameListPage, PlaceListPage, and EventList

    @param: index_list -- a dictionary of either letters or words
    @param: rlocale    -- The locale to use
    @param: current    -- The current page
    @param: rtl        -- Do we use rtl language ?
    @param: only       -- Used only for alphabet letters
    @param: new_page   -- Used to have multiple index pages
    @param: ext        -- The default extension when new_page is used
    """
    # if no letters, return None to its callers
    if not sorted_alpha_index:
        return None

    num_ltrs = len(sorted_alpha_index)
    # num_of_cols = 26
    num_of_cols = num_ltrs
    # num_of_rows = (num_ltrs // num_of_cols) + 1
    num_of_rows = 1

    # begin alphabet navigation division
    with Html("div", id=rtl) as alphabetnavigation:
        index = 0
        output = []
        dup_index = 0
        for dummy_row in range(num_of_rows):
            unordered = Html("ul", class_=rtl)

            cols = 0
            while cols <= num_of_cols and index < num_ltrs:
                menu_item = sorted_alpha_index[index]
                if menu_item == " ":
                    menu_item = "&nbsp;"
                # adding title to hyperlink menu for screen readers and
                # braille writers
                title_txt = "Alphabet Menu: %s" % menu_item
                title_str = rlocale.translation.sgettext(title_txt)
                # deal with multiple ellipsis which are generated for overflow,
                # underflow and inflow labels
                link = menu_item
                if menu_item in output:
                    link = "%s (%i)" % (menu_item, dup_index)
                    dup_index += 1
                output.append(menu_item)

                check_cs = False
                if new_page:
                    new_index, extension = os.path.splitext(new_page)
                    if extension == "":
                        extension = ext
                    pathl = os.path.dirname(new_index)
                    filel = os.path.basename(new_index)
                    if len(filel) > 6:  # 6 is the length of "events"
                        url = "/".join((pathl, "events" + str(index + 1) + extension))
                    else:
                        url = "/".join((pathl, "events" + extension))
                    hyper = Html("a", filel, title=title_str, href=url)
                    if cols != 0:
                        next_name = new_index + str(cols)
                        hyper = Html(
                            "a",
                            menu_item,
                            title=title_str,
                            href="%s" % next_name + extension,
                        )
                    else:
                        next_name = new_index
                        hyper = Html(
                            "a",
                            menu_item,
                            title=title_str,
                            href="%s" % new_index + extension,
                        )
                    current = current.split("_")[0]
                    if current == next_name:
                        check_cs = True
                    temp_cs = 'class = "CurrentSection"'
                    check_cs = temp_cs if check_cs else False
                else:
                    hyper = Html("a", menu_item, title=title_str, href="#%s" % link)
                if check_cs:
                    next_elem = Html("li", hyper, attr=check_cs, inline=True)
                else:
                    next_elem = Html("li", hyper, inline=True)
                unordered.extend(next_elem)

                index += 1
                cols += 1
            num_of_rows -= 1

            alphabetnavigation += unordered

    return alphabetnavigation


def partial_navigation(
    partial_index, rlocale=glocale, current=None, rtl=False, new_page=False, ext=None
):
    """
    Will create the partial navigation bar for big indexes

    @param: partial_index -- a dictionary of (name, handle)
    @param: rlocale       -- The locale to use
    @param: current       -- The current page
    @param: rtl           -- Do we use rtl language ?
    @param: new_page      -- Used to have multiple index pages
    @param: ext           -- The default extension when new_page is used
    """
    # if no letters, return None to its callers
    if not partial_index:
        return None

    num_pages = len(partial_index)
    num_of_rows = 1

    # begin alphabet navigation division
    with Html("div", id=rtl, class_="pnav") as partialnavigation:
        index = 0
        for dummy_row in range(num_of_rows):
            unordered = Html("ul", class_=rtl)

            while index < num_pages:
                handle, name = partial_index[index]
                title_txt = "Go to: "
                title_str = rlocale.translation.sgettext(title_txt) + name
                check_cs = False
                if new_page:
                    if index == 0:
                        new_index = new_page
                    else:
                        new_index = new_page[:-1]
                    new_index, extension = os.path.splitext(new_page)
                    if extension == "":
                        extension = ext
                    if index == 0:
                        next_name = new_index
                        hyper = Html(
                            "a",
                            name + " \u27a1 ",
                            title=title_str,
                            href="%s" % new_index + extension,
                        )
                    else:
                        next_name = new_index + "_%d" % index
                        hyper = Html(
                            "a",
                            name + " \u27a1 ",
                            title=title_str,
                            href="%s" % new_index + "_%d" % index + extension,
                        )
                    if current == next_name:
                        check_cs = True
                    temp_cs = 'class = "CurrentSection"'
                    check_cs = temp_cs if check_cs else False
                else:
                    hyper = Html("a", name, title=title_str, href="#%s" % "link")
                if check_cs:
                    next_elem = Html("li", hyper, attr=check_cs, inline=True)
                else:
                    next_elem = Html("li", hyper, inline=True)
                unordered.extend(next_elem)

                index += 1
            num_of_rows -= 1
            partialnavigation += unordered
    return partialnavigation


def _has_webpage_extension(url):
    """
    determine if a filename has an extension or not...

    @param: url -- filename to be checked
    """
    return url.endswith(_WEB_EXT)


def add_birthdate(dbase, ppl_handle_list, rlocale):
    """
    This will sort a list of child handles in birth order
    For each entry in the list, we'll have :
         birth date
         The transtated birth date for the configured locale
         The transtated death date for the configured locale
         The handle for the child

    @param: dbase           -- The database to use
    @param: ppl_handle_list -- the handle for the people
    @param: rlocale         -- the locale for date translation
    """
    sortable_individuals = []
    for person_handle in ppl_handle_list:
        birth_date = 0  # dummy value in case none is found
        person = dbase.get_person_from_handle(person_handle)
        if person:
            birth_ref = person.get_birth_ref()
            birth1 = ""
            if birth_ref:
                birth = dbase.get_event_from_handle(birth_ref.ref)
                if birth:
                    birth1 = rlocale.get_date(birth.get_date_object())
                    birth_date = birth.get_date_object().get_sort_value()
            death_event = get_death_or_fallback(dbase, person)
            if death_event:
                death = rlocale.get_date(death_event.get_date_object())
            else:
                death = ""
        sortable_individuals.append((birth_date, birth1, death, person_handle))

    # return a list of handles with the individual's birthdate attached
    return sortable_individuals


def _find_birth_date(dbase, individual):
    """
    will look for a birth date within the person's events

    @param: dbase      -- The database to use
    @param: individual -- The individual for who we want to find the birth date
    """
    date_out = None
    birth_ref = individual.get_birth_ref()
    if birth_ref:
        birth = dbase.get_event_from_handle(birth_ref.ref)
        if birth:
            date_out = birth.get_date_object()
            date_out.fallback = False
    else:
        person_evt_ref_list = individual.get_primary_event_ref_list()
        if person_evt_ref_list:
            for evt_ref in person_evt_ref_list:
                event = dbase.get_event_from_handle(evt_ref.ref)
                if event:
                    if event.get_type().is_birth_fallback():
                        date_out = event.get_date_object()
                        date_out.fallback = True
                        LOG.debug("setting fallback to true for '%s'", event)
                        break
    return date_out


def _find_death_date(dbase, individual):
    """
    will look for a death date within a person's events

    @param: dbase      -- The database to use
    @param: individual -- The individual for who we want to find the death date
    """
    date_out = None
    death_ref = individual.get_death_ref()
    if death_ref:
        death = dbase.get_event_from_handle(death_ref.ref)
        if death:
            date_out = death.get_date_object()
            date_out.fallback = False
    else:
        person_evt_ref_list = individual.get_primary_event_ref_list()
        if person_evt_ref_list:
            for evt_ref in person_evt_ref_list:
                event = dbase.get_event_from_handle(evt_ref.ref)
                if event:
                    if event.get_type().is_death_fallback():
                        date_out = event.get_date_object()
                        date_out.fallback = True
                        LOG.debug("setting fallback to true for '%s'", event)
                        break
    return date_out


def build_event_data_by_individuals(dbase, ppl_handle_list):
    """
    creates a list of event handles and event types for this database

    @param: dbase           -- The database to use
    @param: ppl_handle_list -- the handle for the people
    """
    event_handle_list = []
    event_types = []

    for person_handle in ppl_handle_list:
        person = dbase.get_person_from_handle(person_handle)
        if person:
            evt_ref_list = person.get_event_ref_list()
            if evt_ref_list:
                for evt_ref in evt_ref_list:
                    event = dbase.get_event_from_handle(evt_ref.ref)
                    if event:
                        event_types.append(str(event.get_type()))
                        event_handle_list.append(evt_ref.ref)

            person_family_handle_list = person.get_family_handle_list()
            if person_family_handle_list:
                for family_handle in person_family_handle_list:
                    family = dbase.get_family_from_handle(family_handle)
                    if family:
                        family_evt_ref_list = family.get_event_ref_list()
                        if family_evt_ref_list:
                            for evt_ref in family_evt_ref_list:
                                event = dbase.get_event_from_handle(evt_ref.ref)
                                if event:
                                    event_types.append(str(event.type))
                                    event_handle_list.append(evt_ref.ref)

    # return event_handle_list and event types to its caller
    return event_handle_list, event_types


def name_to_md5(text):
    """This creates an MD5 hex string to be used as filename."""

    return md5(text.encode("utf-8")).hexdigest()


def get_gendex_data(database, event_ref):
    """
    Given an event, return the date and place a strings

    @param: database  -- The database
    @param: event_ref -- The event reference
    """
    doe = ""  # date of event
    poe = ""  # place of event
    if event_ref and event_ref.ref:
        event = database.get_event_from_handle(event_ref.ref)
        if event:
            date = event.get_date_object()
            doe = format_date(date)
            if event.get_place_handle():
                place_handle = event.get_place_handle()
                if place_handle:
                    place = database.get_place_from_handle(place_handle)
                    if place:
                        poe = _pd.display(database, place, date)
    return doe, poe


def format_date(date):
    """
    Format the date
    """
    start = date.get_start_date()
    if start != Date.EMPTY:
        cal = date.get_calendar()
        mod = date.get_modifier()
        quality = date.get_quality()
        if quality in DATE_QUALITY:
            qual_text = DATE_QUALITY[quality] + " "
        else:
            qual_text = ""
        if mod == Date.MOD_SPAN:
            val = "%sFROM %s TO %s" % (
                qual_text,
                make_gedcom_date(start, cal, mod, None),
                make_gedcom_date(date.get_stop_date(), cal, mod, None),
            )
        elif mod == Date.MOD_RANGE:
            val = "%sBET %s AND %s" % (
                qual_text,
                make_gedcom_date(start, cal, mod, None),
                make_gedcom_date(date.get_stop_date(), cal, mod, None),
            )
        else:
            val = make_gedcom_date(start, cal, mod, quality)
        return val
    return ""


# This command then defines the 'html_escape' option for escaping
# special characters for presentation in HTML based on the above list.
def html_escape(text):
    """Convert the text and replace some characters with a &# variant."""

    # First single characters, no quotes
    text = escape(text)

    # Deal with double quotes.
    match = _HTML_DBL_QUOTES.match(text)
    while match:
        text = "%s" "&#8220;" "%s" "&#8221;" "%s" % match.groups()
        match = _HTML_DBL_QUOTES.match(text)
    # Replace remaining double quotes.
    text = text.replace('"', "&#34;")

    # Deal with single quotes.
    text = text.replace("'s ", "&#8217;s ")
    match = _HTML_SNG_QUOTES.match(text)
    while match:
        text = "%s" "&#8216;" "%s" "&#8217;" "%s" % match.groups()
        match = _HTML_SNG_QUOTES.match(text)
    # Replace remaining single quotes.
    text = text.replace("'", "&#39;")

    return text


def create_indexes_pages(report, name, index_list, function, handle_list, locale=None):
    """
    This is used to create indexes depending on a row limit.

    @param: report      -- The instance of the main report class
    @param: name        -- The base name of the file to create
    @param: index_list  -- The base name of the file to create
    @param: function    -- The function used to create the page
    @param: handle_list -- The list of handles to manage
    @param: locale      -- The locale used for sor t, ...
    """
    row_count = report.options["splitindex"]
    page = 0
    for letter, hdle_list in handle_list:
        if "events" not in name:
            hdle_list.sort(key=lambda x: locale.sort_key(x[1]))
        if len(hdle_list) <= row_count:
            function(
                report,
                index_list,
                name,
                letter,
                hdle_list,
                part=page,
            )
        else:
            partial_list = []
            for idx in range(0, int(len(hdle_list) / row_count) + 1, 1):
                if (idx * row_count) < len(hdle_list):
                    partial_list.append(hdle_list[idx * row_count])
            sub_page = 0
            part = 0
            start = 0
            while start < len(hdle_list):
                phdle_list = []
                for nbh in range(len(hdle_list)):
                    if nbh < start:
                        continue
                    if nbh >= start + row_count:
                        break
                    if nbh > len(hdle_list):
                        break
                    phdle_list.append(hdle_list[nbh])
                if sub_page != 0:
                    subp = "_%d" % sub_page
                else:
                    subp = None
                function(
                    report,
                    index_list,
                    name,
                    letter,
                    phdle_list,
                    part=page,
                    subp=subp,
                    partial_list=partial_list,
                )
                part += 1
                start += row_count
                sub_page += 1
        page += 1



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
# Copyright (C) 2017       Paul Franklin
# Copyright (c) 2020       Steve Youngs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Date parsing class. Serves as the base class for any localized
date parsing class. The default base class provides parsing for English.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
import re
import calendar

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".DateParser")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date, DateError, Today
from ..const import GRAMPS_LOCALE as glocale
from ..utils.grampslocale import GrampsLocale
from ._datestrings import DateStrings

# -------------------------------------------------------------------------
#
# Top-level module functions
#
# -------------------------------------------------------------------------
_max_days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
_leap_days = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]


def gregorian_valid(date_tuple):
    """Checks if date_tuple is a valid date in Gregorian Calendar"""
    day = date_tuple[0]
    month = date_tuple[1]
    valid = True
    try:
        if month > 12:
            valid = False
        elif calendar.isleap(date_tuple[2]):
            if day > _leap_days[month - 1]:
                valid = False
        elif day > _max_days[month - 1]:
            valid = False
    except:
        valid = False
    return valid


def julian_valid(date_tuple):
    """Checks if date_tuple is a valid date in Julian Calendar"""
    day = date_tuple[0]
    month = date_tuple[1]
    valid = True
    try:
        if month > 12:
            valid = False
        elif (date_tuple[2]) % 4 == 0:
            # julian always have leapyear every 4 year
            if day > _leap_days[month - 1]:
                valid = False
        elif day > _max_days[month - 1]:
            valid = False
    except:
        valid = False
    return valid


def swedish_valid(date_tuple):
    """Checks if date_tuple is a valid date in Swedish Calendar"""
    valid_J = julian_valid(date_tuple)
    date_tuple = (date_tuple[2], date_tuple[1], date_tuple[0])
    # Swedish calendar starts as Julian 1700-03-01 and ends 1712-03-01 as Julian
    if date_tuple >= (1700, 2, 29) and date_tuple < (1712, 3, 1):
        if date_tuple == (1712, 2, 30):  # extra day was inserted 1712, not valid Julian
            return True
        if valid_J:
            if date_tuple == (1700, 2, 29):  # leapday 1700 was skipped
                return False
            return True
        else:
            return False
    else:
        return False


def french_valid(date_tuple):
    """Checks if date_tuple is a valid date in French Calendar"""
    valid = True
    # year 1 starts on 22.9.1792
    if date_tuple[2] < 1:
        valid = False
    return valid


def _build_prefix_table(month_to_int, month_variants):
    """
    Populate a DateParser.month_to_int-like dict
    with all the prefixes found in month_variants.
    """

    month_variants = list(month_variants)  # drain the generator, if any
    month_to_int_new = {}

    # Populate with full names first, w/o prefixes
    log.debug("Mapping full names...")
    for i in range(0, len(month_variants)):
        for month in month_variants[i]:
            m = month.lower()
            log.debug("Mapping {} -> {}".format(m, i))
            month_to_int_new[m] = i
    month_to_int.update(month_to_int_new)

    log.debug("Mapping new prefixes...")
    months_sorted = list(month_to_int_new.keys())
    months_sorted.sort(key=len, reverse=True)
    for m in months_sorted:
        for prefixlen in reversed(range(1, len(m))):
            mp = m[:prefixlen]
            if mp.strip() != mp:
                continue
            if mp in month_to_int:
                break
            else:
                i = month_to_int[m]
                log.debug("Mapping {} -> {}".format(mp, i))
                month_to_int[mp] = i


def _generate_variants(months):
    """
    Generate all month variants for passing to _build_prefix_table
    :param months: an iterable ordered collection, 1st item is empty, the rest
                   1..N, for a calendar with N months overall, contain, each,
                   an iterable of alternative specifications.
                   Each such specification can be:
                     1) a Lexeme, supporting .variants() to return the list of
                        variants underneath
                     2) a literal string
                     3) a |-separated string of alternatives
                   Empty strings are discarded.
    :return: generator of lists per month with all variants listed once only
             the 1st item will be empty
    """

    for month_lexemes_and_alternatives in months:
        v = []
        for m in month_lexemes_and_alternatives:
            try:
                # Lexeme? ask it to compute the variants it knows
                mv = list(m.variants())
            except AttributeError:
                # plain string, not a lexeme with inflections...
                # Maybe a '|'-separated list of alternatives, maybe empty,
                # maybe a single string. Suppress empty strings!
                mv = (s for s in m.split("|") if s)
            v.extend(mv)
        yield (list(set(v)))


# -------------------------------------------------------------------------
#
# DateParser class
#
# -------------------------------------------------------------------------
class DateParser:
    """
    Convert a text string into a :class:`.Date` object. If the date cannot be
    converted, the text string is assigned.
    """

    _dhformat_parse = re.compile(r".*%(\S).*%(\S).*%(\S).*")

    # RFC-2822 only uses capitalized English abbreviated names, no locales.
    _rfc_days = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat")
    _rfc_mons_to_int = {
        "Jan": 1,
        "Feb": 2,
        "Mar": 3,
        "Apr": 4,
        "May": 5,
        "Jun": 6,
        "Jul": 7,
        "Aug": 8,
        "Sep": 9,
        "Oct": 10,
        "Nov": 11,
        "Dec": 12,
    }

    # seeded with __init_prefix_tables
    month_to_int: dict[str, int]
    swedish_to_int = month_to_int = {}
    """
    Map Gregorian month names and their prefixes, wherever unambiguous,
    to the relevant integer index (1..12).
    """

    # modifiers before the date
    # (overridden if a locale-specific date parser exists)
    modifier_to_int = {
        "before": Date.MOD_BEFORE,
        "bef": Date.MOD_BEFORE,
        "bef.": Date.MOD_BEFORE,
        "after": Date.MOD_AFTER,
        "aft": Date.MOD_AFTER,
        "aft.": Date.MOD_AFTER,
        "about": Date.MOD_ABOUT,
        "abt.": Date.MOD_ABOUT,
        "abt": Date.MOD_ABOUT,
        "circa": Date.MOD_ABOUT,
        "c.": Date.MOD_ABOUT,
        "around": Date.MOD_ABOUT,
        "from": Date.MOD_FROM,
        "to": Date.MOD_TO,
    }
    # in some languages some of above listed modifiers are after the date,
    # in that case the subclass should put them into this dictionary instead
    modifier_after_to_int: dict[str, int] = {}

    hebrew_to_int = {
        "tishri": 1,
        "heshvan": 2,
        "kislev": 3,
        "tevet": 4,
        "shevat": 5,
        "adari": 6,
        "adarii": 7,
        "nisan": 8,
        "iyyar": 9,
        "sivan": 10,
        "tammuz": 11,
        "av": 12,
        "elul": 13,
        # alternative spelling
        "cheshvan": 2,
        "adar sheni": 7,
        "iyar": 9,
        # GEDCOM months
        "tsh": 1,
        "csh": 5,
        "ksl": 3,
        "tvt": 4,
        "shv": 5,
        "adr": 6,
        "ads": 7,
        "nsn": 8,
        "iyr": 9,
        "svn": 10,
        "tmz": 11,
        "aav": 12,
        "ell": 13,
    }

    french_to_int = {
        # the long ones are seeded with __init_prefix_tables
        # GEDCOM months
        "vend": 1,
        "brum": 2,
        "frim": 3,
        "nivo": 4,
        "pluv": 5,
        "vent": 6,
        "germ": 7,
        "flor": 8,
        "prai": 9,
        "mess": 10,
        "ther": 11,
        "fruc": 12,
        "comp": 13,
    }

    islamic_to_int = {
        # some are already seeded with __init_prefix_tables,
        # but it is a pain to separate them out from the variants...
        "muharram": 1,
        "muharram ul haram": 1,
        "safar": 2,
        "rabi`al-awwal": 3,
        "rabi'l": 3,
        "rabi`ul-akhir": 4,
        "rabi`ath-thani": 4,
        "rabi` ath-thani": 4,
        "rabi`al-thaany": 4,
        "rabi` al-thaany": 4,
        "rabi' ii": 4,
        "jumada l-ula": 5,
        "jumaada-ul-awwal": 5,
        "jumaada i": 5,
        "jumada t-tania": 6,
        "jumaada-ul-akhir": 6,
        "jumaada al-thaany": 6,
        "jumaada ii": 5,
        "rajab": 7,
        "sha`ban": 8,
        "sha`aban": 8,
        "ramadan": 9,
        "ramadhan": 9,
        "shawwal": 10,
        "dhu l-qa`da": 11,
        "dhu qadah": 11,
        "thw al-qi`dah": 11,
        "dhu l-hijja": 12,
        "dhu hijja": 12,
        "thw al-hijjah": 12,
    }

    # seeded with __init_prefix_tables
    persian_to_int: dict[str, int] = {}

    bce = ["B.C.E.", "B.C.E", "BCE", "B.C.", "B.C", "BC"]
    # (overridden if a locale-specific date parser exists)

    # seeded with __init_prefix_tables
    calendar_to_int: dict[str, int] = {}
    # (probably overridden if a locale-specific date parser exists)

    newyear_to_int = {
        "jan1": Date.NEWYEAR_JAN1,
        "mar1": Date.NEWYEAR_MAR1,
        "mar25": Date.NEWYEAR_MAR25,
        "sep1": Date.NEWYEAR_SEP1,
    }

    quality_to_int = {
        "estimated": Date.QUAL_ESTIMATED,
        "est.": Date.QUAL_ESTIMATED,
        "est": Date.QUAL_ESTIMATED,
        "calc.": Date.QUAL_CALCULATED,
        "calc": Date.QUAL_CALCULATED,
        "calculated": Date.QUAL_CALCULATED,
    }
    # (overridden if a locale-specific date parser exists)

    today = [
        "$T",
    ]
    # Override with a list of *synonyms* for "today" in your language.
    # Note: the word "today" itself will already be pulled in from your translation DB,
    # see init_strings, so there is no need to override this if you have no aliases
    # for "today".
    # We also secretly support "$T" like in some reports.

    _langs: set[str] = set()

    def __init_prefix_tables(self):
        ds = self._ds = DateStrings(self._locale)
        lang = self._locale.lang
        if lang in DateParser._langs:
            log.debug("Prefix tables for {} already built".format(lang))
            return
        else:
            DateParser._langs.add(lang)
        log.debug("Begin building parser prefix tables for {}".format(lang))
        _build_prefix_table(
            DateParser.month_to_int,
            _generate_variants(
                zip(ds.long_months, ds.short_months, ds.swedish_SV, ds.alt_long_months)
            ),
        )
        _build_prefix_table(
            DateParser.hebrew_to_int, _generate_variants(zip(ds.hebrew))
        )
        _build_prefix_table(
            DateParser.french_to_int, _generate_variants(zip(ds.french))
        )
        _build_prefix_table(
            DateParser.islamic_to_int, _generate_variants(zip(ds.islamic))
        )
        _build_prefix_table(
            DateParser.persian_to_int, _generate_variants(zip(ds.persian))
        )
        _build_prefix_table(
            DateParser.calendar_to_int, _generate_variants(zip(ds.calendar))
        )

    def __init__(self, plocale=None):
        """
        :param plocale: allow correct date format to be loaded and parsed
        :type plocale: a :class:`.GrampsLocale` instance
        """
        from ._datehandler import locale_tformat  # circular import if above

        if plocale:
            self._locale = plocale
        elif not hasattr(self, "_locale"):
            self._locale = glocale
        if self._locale.calendar in locale_tformat:
            self.dhformat = locale_tformat[self._locale.calendar]  # date format
        else:
            self.dhformat = locale_tformat["en_GB"]  # something is required
        self.dhformat_changed()  # Allow overriding so a subclass can modify it
        self.init_strings()
        self.parser = {
            Date.CAL_GREGORIAN: self._parse_gregorian,
            Date.CAL_JULIAN: self._parse_julian,
            Date.CAL_FRENCH: self._parse_french,
            Date.CAL_PERSIAN: self._parse_persian,
            Date.CAL_HEBREW: self._parse_hebrew,
            Date.CAL_ISLAMIC: self._parse_islamic,
            Date.CAL_SWEDISH: self._parse_swedish,
        }

        match = self._dhformat_parse.match(self.dhformat.lower())
        if match:
            self._ddmy = match.groups() == ("a", "e", "b", "y")  # Icelandic
            self.dmy = (
                match.groups() == ("d", "m", "y")
                or match.groups() == ("e", "m", "y")
                or match.groups() == ("d", "b", "y")  # Bulgarian
                or match.groups() == ("a", "e", "b", "y")
            )
            self.ymd = match.groups() == ("y", "m", "d") or match.groups() == (
                "y",
                "b",
                "d",
            )
            # note ('m','d','y') is valid -- in which case both will be False
        else:
            self.dmy = True
            self.ymd = False
            self._ddmy = False

    def dhformat_changed(self):
        """Allow overriding so a subclass can modify it"""
        pass

    def re_longest_first(self, keys):
        """
        returns a string for a RE group which contains the given keys
        sorted so that longest keys match first.  Any '.' characters
        are quoted.
        """
        keys.sort(key=len, reverse=True)
        return "(" + "|".join([re.escape(key) for key in keys]) + ")"

    def init_strings(self):
        """
        This method compiles regular expression strings for matching dates.

        Most of the re's in most languages can stay as is. span and range
        most likely will need to change. Whatever change is done, this method
        may be called first as DateParser.init_strings(self) so that the
        invariant expresions don't need to be repeatedly coded. All differences
        can be coded after DateParser.init_strings(self) call, that way they
        override stuff from this method.

        .. seealso:: :class:`.DateParserRU` as an example.
        """
        _ = self._locale.translation.gettext
        self.__init_prefix_tables()

        self._rfc_mon_str = "(" + "|".join(list(self._rfc_mons_to_int.keys())) + ")"
        self._rfc_day_str = "(" + "|".join(self._rfc_days) + ")"

        self._bce_str = self.re_longest_first(self.bce)
        self._qual_str = self.re_longest_first(list(self.quality_to_int.keys()))
        self._mod_str = self.re_longest_first(list(self.modifier_to_int.keys()))
        self._mod_after_str = self.re_longest_first(
            list(self.modifier_after_to_int.keys())
        )

        self._mon_str = self.re_longest_first(list(self.month_to_int.keys()))
        self._jmon_str = self.re_longest_first(list(self.hebrew_to_int.keys()))
        self._fmon_str = self.re_longest_first(list(self.french_to_int.keys()))
        self._pmon_str = self.re_longest_first(list(self.persian_to_int.keys()))
        self._imon_str = self.re_longest_first(list(self.islamic_to_int.keys()))
        self._smon_str = self.re_longest_first(list(self.swedish_to_int.keys()))
        self._cal_str = self.re_longest_first(list(self.calendar_to_int.keys()))
        self._ny_str = self.re_longest_first(list(self.newyear_to_int.keys()))

        self._today_str = self.re_longest_first(
            self.today
            + [
                _("today"),
            ]
        )

        # bce, calendar type and quality may be either at the end or at
        # the beginning of the given date string, therefore they will
        # be parsed from the middle and will be in match.group(2).
        self._bce_re = re.compile(r"(.*)\s+%s( ?.*)" % self._bce_str)

        self._cal = re.compile(r"(.*)\s+\(%s\)( ?.*)" % self._cal_str, re.IGNORECASE)
        self._calny = re.compile(
            r"(.*)\s+\(%s,\s*%s\)( ?.*)" % (self._cal_str, self._ny_str), re.IGNORECASE
        )
        self._calny_iso = re.compile(
            r"(.*)\s+\(%s,\s*(\d{1,2}-\d{1,2})\)( ?.*)" % self._cal_str, re.IGNORECASE
        )

        self._ny = re.compile(r"(.*)\s+\(%s\)( ?.*)" % self._ny_str, re.IGNORECASE)
        self._ny_iso = re.compile(r"(.*)\s+\((\d{1,2}-\d{1,2})\)( ?.*)")

        self._qual = re.compile(r"(.* ?)%s\s+(.+)" % self._qual_str, re.IGNORECASE)

        self._span = re.compile(
            r"(from)\s+(?P<start>.+)\s+to\s+(?P<stop>.+)", re.IGNORECASE
        )
        self._range = re.compile(
            r"(bet|bet.|between)\s+(?P<start>.+)\s+and\s+(?P<stop>.+)", re.IGNORECASE
        )
        self._quarter = re.compile(
            r"[qQ](?P<quarter>[1-4])\s+(?P<year>.+)", re.IGNORECASE
        )
        self._modifier = re.compile(r"%s\s+(.*)" % self._mod_str, re.IGNORECASE)
        self._modifier_after = re.compile(
            r"(.*)\s+%s" % self._mod_after_str, re.IGNORECASE
        )
        self._abt2 = re.compile("<(.*)>", re.IGNORECASE)
        self._text = re.compile(
            r"%s\.?(\s+\d+)?\s*,?\s+((\d+)(/\d+)?)?\s*$" % self._mon_str, re.IGNORECASE
        )
        # this next RE has the (possibly-slashed) year at the string's end
        self._text2 = re.compile(
            r"(\d+)?\s+?%s\.?\s*((\d+)(/\d+)?)?\s*$" % self._mon_str, re.IGNORECASE
        )
        self._jtext = re.compile(
            r"%s\s+(\d+)?\s*,?\s*((\d+)(/\d+)?)?\s*$" % self._jmon_str, re.IGNORECASE
        )
        self._jtext2 = re.compile(
            r"(\d+)?\s+?%s\s*((\d+)(/\d+)?)?\s*$" % self._jmon_str, re.IGNORECASE
        )
        self._ftext = re.compile(
            r"%s\s+(\d+)?\s*,?\s*((\d+)(/\d+)?)?\s*$" % self._fmon_str, re.IGNORECASE
        )
        self._ftext2 = re.compile(
            r"(\d+)?\s+?%s\s*((\d+)(/\d+)?)?\s*$" % self._fmon_str, re.IGNORECASE
        )
        self._ptext = re.compile(
            r"%s\s+(\d+)?\s*,?\s*((\d+)(/\d+)?)?\s*$" % self._pmon_str, re.IGNORECASE
        )
        self._ptext2 = re.compile(
            r"(\d+)?\s+?%s\s*((\d+)(/\d+)?)?\s*$" % self._pmon_str, re.IGNORECASE
        )
        self._itext = re.compile(
            r"%s\s+(\d+)?\s*,?\s*((\d+)(/\d+)?)?\s*$" % self._imon_str, re.IGNORECASE
        )
        self._itext2 = re.compile(
            r"(\d+)?\s+?%s\s*((\d+)(/\d+)?)?\s*$" % self._imon_str, re.IGNORECASE
        )
        self._stext = re.compile(
            r"%s\.?\s+(\d+)?\s*,?\s*((\d+)(/\d+)?)?\s*$" % self._smon_str, re.IGNORECASE
        )
        self._stext2 = re.compile(
            r"(\d+)?\s+?%s\.?\s*((\d+)(/\d+)?)?\s*$" % self._smon_str, re.IGNORECASE
        )
        self._numeric = re.compile(r"((\d+)[/\.]\s*)?((\d+)[/\.]\s*)?(\d+)\s*$")
        self._iso = re.compile(r"(\d+)(/(\d+))?-(\d+)(-(\d+))?\s*$")
        self._isotimestamp = re.compile(
            r"^\s*?(\d{4})([01]\d)([0123]\d)(?:(?:[012]\d[0-5]\d[0-5]\d)|"
            r"(?:\s+[012]\d:[0-5]\d(?::[0-5]\d)?))?\s*?$"
        )
        self._rfc = re.compile(
            r"(%s,)?\s+(\d|\d\d)\s+%s\s+(\d+)\s+\d\d:\d\d(:\d\d)?\s+"
            r"(\+|-)\d\d\d\d" % (self._rfc_day_str, self._rfc_mon_str)
        )
        self._today = re.compile(r"^\s*%s\s*$" % self._today_str, re.IGNORECASE)

    def _get_int(self, val):
        """
        Convert the string to an integer if the value is not None. If the
        value is None, a zero is returned
        """
        if val is None:
            return 0
        else:
            return int(val)

    def _parse_hebrew(self, text):
        return self._parse_calendar(text, self._jtext, self._jtext2, self.hebrew_to_int)

    def _parse_islamic(self, text):
        return self._parse_calendar(
            text, self._itext, self._itext2, self.islamic_to_int
        )

    def _parse_persian(self, text):
        return self._parse_calendar(
            text, self._ptext, self._ptext2, self.persian_to_int
        )

    def _parse_french(self, text):
        return self._parse_calendar(
            text, self._ftext, self._ftext2, self.french_to_int, french_valid
        )

    def _parse_gregorian(self, text):
        return self._parse_calendar(
            text, self._text, self._text2, self.month_to_int, gregorian_valid
        )

    def _parse_julian(self, text):
        return self._parse_calendar(
            text, self._text, self._text2, self.month_to_int, julian_valid
        )

    def _parse_swedish(self, text):
        return self._parse_calendar(
            text, self._stext, self._stext2, self.swedish_to_int, swedish_valid
        )

    def _parse_calendar(self, text, regex1, regex2, mmap, check=None):
        match = regex1.match(text.lower())
        if match:  # user typed in 'month-name day year' or 'month-name year'
            groups = match.groups()
            if groups[0] is None:
                m = 0
            else:
                m = mmap[groups[0].lower()]

            if groups[2] is None:
                y = self._get_int(groups[1])
                d = 0
                s = False
            else:
                d = self._get_int(groups[1])
                if groups[4] is not None:
                    y = int(groups[3]) + 1  # fullyear + 1
                    s = True  # slash year
                else:  # regular year
                    y = int(groups[3])
                    s = False
            value = (d, m, y, s)
            if s and julian_valid(value):  # slash year
                pass
            elif check and not check((d, m, y)):
                value = Date.EMPTY
            return value

        match = regex2.match(text.lower())
        if match:  # user typed in day month-name year or year month-name day
            groups = match.groups()
            if self.ymd:
                if groups[3] is None:
                    m = 0
                else:
                    m = mmap[groups[3].lower()]
                d = self._get_int(groups[4])
                if groups[0] is None:
                    y = None
                    s = False
                else:
                    if groups[2] is not None:
                        y = int(groups[1]) + 1  # fullyear + 1
                        s = True  # slash year
                    else:  # regular year
                        y = int(groups[1])
                        s = False
            else:
                if groups[1] is None:
                    m = 0
                else:
                    m = mmap[groups[1].lower()]
                d = self._get_int(groups[0])
                if groups[2] is None:
                    y = None
                    s = False
                else:
                    if groups[4] is not None:
                        y = int(groups[3]) + 1  # fullyear + 1
                        s = True  # slash year
                    else:  # regular year
                        y = int(groups[3])
                        s = False
            value = (d, m, y, s)
            if check and not check((d, m, y)):
                value = Date.EMPTY
            return value

        return Date.EMPTY

    def _parse_subdate(self, text, subparser=None, cal=None):
        """
        Convert only the date portion of a date.
        """
        if subparser is None:
            subparser = self._parse_gregorian

        if subparser == self._parse_gregorian:
            check = gregorian_valid

        elif subparser == self._parse_julian:
            check = julian_valid

        elif subparser == self._parse_swedish:
            check = swedish_valid

        elif subparser == self._parse_french:
            check = french_valid
        else:
            check = None

        value = subparser(text)
        if value != Date.EMPTY:
            return value

        match = self._iso.match(text)
        if match:
            groups = match.groups()
            y = self._get_int(groups[0])
            m = self._get_int(groups[3])
            d = self._get_int(groups[5])
            if groups[2] and julian_valid((d, m, y + 1)):
                return (d, m, y + 1, True)  # slash year
            if check is None or check((d, m, y)):
                return (d, m, y, False)
            return Date.EMPTY

        # Database datetime format, used in ex. MSSQL
        # YYYYMMDD HH:MM:SS or YYYYMMDD or YYYYMMDDHHMMSS
        match = self._isotimestamp.match(text)
        if match:
            groups = match.groups()
            y = self._get_int(groups[0])
            m = self._get_int(groups[1])
            d = self._get_int(groups[2])
            value = (d, m, y, False)
            if not check((d, m, y)):
                value = Date.EMPTY
            return value

        match = self._rfc.match(text)
        if match:
            groups = match.groups()
            d = self._get_int(groups[2])
            m = self._rfc_mons_to_int[groups[3]]
            y = self._get_int(groups[4])
            value = (d, m, y, False)
            if check and not check((d, m, y)):
                value = Date.EMPTY
            return value

        match = self._numeric.match(text)
        if match:
            groups = match.groups()
            if groups == (None, None, None, None, None):
                return Date.EMPTY
            if self._ddmy:  # Icelandic
                groups = groups[1:]  # ignore the day of the week at the start
            if self.ymd:
                # '1789' and ymd: incomplete date
                if groups[1] is None:
                    y = self._get_int(groups[4])
                    m = 0
                    d = 0
                elif groups[3] is None:
                    y = self._get_int(groups[1])
                    m = self._get_int(groups[4])
                    d = 0
                else:
                    y = self._get_int(groups[1])
                    m = self._get_int(groups[3])
                    d = self._get_int(groups[4])
                if m > 12:  # maybe a slash year, not a month (1722/3 is March)
                    if y % 100 == 99:
                        modyear = (y + 1) % 1000
                    elif y % 10 == 9:
                        modyear = (y + 1) % 100
                    else:
                        modyear = (y + 1) % 10
                    if m == modyear:
                        return (0, 0, y + 1, True)  # slash year
            else:
                y = self._get_int(groups[4])
                if self.dmy:
                    if groups[3] is None:
                        m = self._get_int(groups[1])
                        d = 0
                    else:
                        m = self._get_int(groups[3])
                        d = self._get_int(groups[1])
                else:
                    m = self._get_int(groups[1])
                    d = self._get_int(groups[3])
                if m > 12:  # maybe a slash year, not a month
                    if m % 100 == 99:
                        modyear = (m + 1) % 1000
                    elif m % 10 == 9:
                        modyear = (m + 1) % 100
                    else:
                        modyear = (m + 1) % 10
                    if y == modyear:
                        return (0, 0, m + 1, True)  # slash year
            value = (d, m, y, False)
            if check and not check((d, m, y)):
                value = Date.EMPTY
            return value

        match = self._today.match(text)
        if match:
            today = Today()
            if cal:
                today = today.to_calendar(cal)
            return today.get_dmy(get_slash=True)

        return Date.EMPTY

    def match_calendar(self, text, cal):
        """
        Try parsing calendar.

        Return calendar index and the text with calendar removed.
        """
        match = self._cal.match(text)
        if match:
            cal = self.calendar_to_int[match.group(2).lower()]
            text = match.group(1) + match.group(3)
        return (text, cal)

    def match_calendar_newyear(self, text, cal, newyear):
        """
        Try parsing calendar and newyear code.

        Return newyear index and the text with calendar removed.
        """
        match = self._calny.match(text)
        if match:
            cal = self.calendar_to_int[match.group(2).lower()]
            newyear = self.newyear_to_int[match.group(3).lower()]
            text = match.group(1) + match.group(4)
        else:
            match = self._calny_iso.match(text)
            if match:
                cal = self.calendar_to_int[match.group(2).lower()]
                newyear = tuple(map(int, match.group(3).split("-")))
                text = match.group(1) + match.group(4)
        return (text, cal, newyear)

    def match_newyear(self, text, newyear):
        """
        Try parsing calendar and newyear code.

        Return newyear index and the text with calendar removed.
        """
        match = self._ny.match(text)
        if match:
            newyear = self.newyear_to_int[match.group(2).lower()]
            text = match.group(1) + match.group(3)
        else:
            match = self._ny_iso.match(text)
            if match:
                newyear = tuple(map(int, match.group(2).split("-")))
                text = match.group(1) + match.group(3)
        return (text, newyear)

    def match_quality(self, text, qual):
        """
        Try matching quality.

        Return quality index and the text with quality removed.
        """
        match = self._qual.match(text)
        if match:
            qual = self.quality_to_int[match.group(2).lower()]
            text = match.group(1) + match.group(3)
        return (text, qual)

    def match_span(self, text, cal, ny, qual, date):
        """
        Try matching span date.

        On success, set the date and return 1. On failure return 0.
        """
        match = self._span.match(text)
        if match:
            text_parser = self.parser[cal]
            (text1, bc1) = self.match_bce(match.group("start"))
            start = self._parse_subdate(text1, text_parser, cal)
            if start == Date.EMPTY and text1 != "":
                return 0
            if bc1:
                start = self.invert_year(start)

            (text2, bc2) = self.match_bce(match.group("stop"))
            stop = self._parse_subdate(text2, text_parser, cal)
            if stop == Date.EMPTY and text2 != "":
                return 0
            if bc2:
                stop = self.invert_year(stop)

            date.set(qual, Date.MOD_SPAN, cal, start + stop, newyear=ny)
            return 1
        return 0

    def match_range(self, text, cal, ny, qual, date):
        """
        Try matching range date.

        On success, set the date and return 1. On failure return 0.
        """
        match = self._range.match(text)
        if match:
            text_parser = self.parser[cal]
            (text1, bc1) = self.match_bce(match.group("start"))
            start = self._parse_subdate(text1, text_parser, cal)
            if start == Date.EMPTY and text1 != "":
                return 0
            if bc1:
                start = self.invert_year(start)

            (text2, bc2) = self.match_bce(match.group("stop"))
            stop = self._parse_subdate(text2, text_parser, cal)
            if stop == Date.EMPTY and text2 != "":
                return 0
            if bc2:
                stop = self.invert_year(stop)

            date.set(qual, Date.MOD_RANGE, cal, start + stop, newyear=ny)
            return 1
        return 0

    def match_quarter(self, text, cal, ny, qual, date):
        """
        Try matching calendar quarter date.

        On success, set the date and return 1. On failure return 0.
        """
        match = self._quarter.match(text)
        if match:
            quarter = self._get_int(match.group("quarter"))

            text_parser = self.parser[cal]
            (text, bc) = self.match_bce(match.group("year"))
            start = self._parse_subdate(text, text_parser, cal)
            if (
                (start == Date.EMPTY and text != "")
                or (start[0] != 0)
                or (start[1] != 0)
            ):  # reject dates where the day or month have been set
                return 0
            if bc:
                start = self.invert_year(start)

            stop_month = quarter * 3
            stop_day = _max_days[
                stop_month - 1
            ]  # no need to worry about leap years since no quarter ends in February

            date.set(
                qual,
                Date.MOD_RANGE,
                cal,
                (1, stop_month - 2, start[2], start[3])
                + (stop_day, stop_month, start[2], start[3]),
                newyear=ny,
            )
            return 1
        return 0

    def match_bce(self, text):
        """
        Try matching BCE qualifier.

        Return BCE (True/False) and the text with matched part removed.
        """
        match = self._bce_re.match(text)
        bc = False
        if match:
            # bce is in the match.group(2)
            try:
                text = match.group(1) + match.group(3)
            except:
                print("ERROR MATCH:", match.groups())
            bc = True
        return (text, bc)

    def match_modifier(self, text, cal, ny, qual, bc, date):
        """
        Try matching date with modifier.

        On success, set the date and return 1. On failure return 0.
        """
        # modifiers before the date
        match = self._modifier.match(text)
        if match:
            grps = match.groups()
            start = self._parse_subdate(grps[1], self.parser[cal], cal)
            mod = self.modifier_to_int.get(grps[0].lower(), Date.MOD_NONE)
            if start == Date.EMPTY:
                date.set_modifier(Date.MOD_TEXTONLY)
                date.set_text_value(text)
            elif bc:
                date.set(qual, mod, cal, self.invert_year(start), newyear=ny)
            else:
                date.set(qual, mod, cal, start, newyear=ny)
            return True
        # modifiers after the date
        if self.modifier_after_to_int:
            match = self._modifier_after.match(text)
            if match:
                grps = match.groups()
                start = self._parse_subdate(grps[0], self.parser[cal], cal)
                mod = self.modifier_after_to_int.get(grps[1].lower(), Date.MOD_NONE)
                if start == Date.EMPTY:
                    date.set_modifier(Date.MOD_TEXTONLY)
                    date.set_text_value(text)
                elif bc:
                    date.set(qual, mod, cal, self.invert_year(start), newyear=ny)
                else:
                    date.set(qual, mod, cal, start, newyear=ny)
                return True
        match = self._abt2.match(text)
        if match:
            grps = match.groups()
            start = self._parse_subdate(grps[0], cal=cal)
            mod = Date.MOD_ABOUT
            if start == Date.EMPTY:
                date.set_modifier(Date.MOD_TEXTONLY)
                date.set_text_value(text)
            elif bc:
                date.set(qual, mod, cal, self.invert_year(start), newyear=ny)
            else:
                date.set(qual, mod, cal, start, newyear=ny)
            return True
        return False

    def set_date(self, date, text):
        """
        Parses the text and sets the date according to the parsing.
        """
        text = text.strip()  # otherwise spaces can make it a bad date
        date.set_text_value(text)
        qual = Date.QUAL_NONE
        cal = Date.CAL_GREGORIAN
        newyear = Date.NEWYEAR_JAN1

        (text, cal, newyear) = self.match_calendar_newyear(text, cal, newyear)
        (text, newyear) = self.match_newyear(text, newyear)
        (text, cal) = self.match_calendar(text, cal)
        (text, qual) = self.match_quality(text, qual)

        if self.match_span(text, cal, newyear, qual, date):
            return
        if self.match_range(text, cal, newyear, qual, date):
            return
        if self.match_quarter(text, cal, newyear, qual, date):
            return

        (text, bc) = self.match_bce(text)
        if self.match_modifier(text, cal, newyear, qual, bc, date):
            return

        try:
            subdate = self._parse_subdate(text, self.parser[cal], cal)
            if subdate == Date.EMPTY and text != "":
                date.set_as_text(text)
                return
        except:
            date.set_as_text(text)
            return

        if bc:
            date.set(
                qual, Date.MOD_NONE, cal, self.invert_year(subdate), newyear=newyear
            )
        else:
            date.set(qual, Date.MOD_NONE, cal, subdate, newyear=newyear)

    def invert_year(self, subdate):
        return (subdate[0], subdate[1], -subdate[2], subdate[3])

    def parse(self, text):
        """
        Parses the text, returning a :class:`.Date` object.
        """
        new_date = Date()
        try:
            self.set_date(new_date, text)
        except DateError:
            new_date.set_as_text(text)
        return new_date



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2009       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This module provides the base class for plugins.
"""


class Plugin:
    """
    This class serves as a base class for all plugins that can be registered
    with the plugin manager
    """

    def __init__(self, name, description, module_name):
        """
        :param name: A friendly name to call this plugin.
            Example: "GEDCOM Import"
        :type name: string
        :param description: A short description of the plugin.
            Example: "This plugin will import a GEDCOM file into a database"
        :type description: string
        :param module_name: The name of the module that contains this plugin.
            Example: "gedcom"
        :type module_name: string
        :return: nothing
        """
        self.__name = name
        self.__desc = description
        self.__mod_name = module_name

    def get_name(self):
        """
        Get the name of this plugin.

        :return: a string representing the name of the plugin
        """
        return self.__name

    def get_description(self):
        """
        Get the description of this plugin.

        :return: a string that describes the plugin
        """
        return self.__desc

    def get_module_name(self):
        """
        Get the name of the module that this plugin lives in.

        :return: a string representing the name of the module for this plugin
        """
        return self.__mod_name



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


"""
Provide merge capabilities for repositories.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.const import URL_MANUAL_SECT3
from ..display import display_help
from ..managedwindow import ManagedWindow
from gramps.gen.merge import MergeRepositoryQuery

# -------------------------------------------------------------------------
#
# Gramps constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = URL_MANUAL_SECT3
WIKI_HELP_SEC = _("Merge_Repositories", "manual")
_GLADE_FILE = "mergerepository.glade"
REPO_NAME = _("Name:", "repo")


# -------------------------------------------------------------------------
#
# MergeRepository
#
# -------------------------------------------------------------------------
class MergeRepository(ManagedWindow):
    """
    Displays a dialog box that allows two repositories to be combined into one.
    """

    def __init__(self, dbstate, uistate, track, handle1, handle2):
        ManagedWindow.__init__(self, uistate, track, self.__class__)
        self.dbstate = dbstate
        database = dbstate.db
        self.rp1 = database.get_repository_from_handle(handle1)
        self.rp2 = database.get_repository_from_handle(handle2)

        self.define_glade("mergerepository", _GLADE_FILE)
        self.set_window(
            self._gladeobj.toplevel,
            self.get_widget("repository_title"),
            _("Merge Repositories"),
        )
        self.setup_configs("interface.merge-repository", 500, 250)

        # Detailed selection widgets
        name1 = self.rp1.get_name()
        name2 = self.rp2.get_name()
        for widget_name in ("name_btn1", "name_btn2"):
            self.get_widget(widget_name).set_label(REPO_NAME)
        entry1 = self.get_widget("name1")
        entry2 = self.get_widget("name2")
        entry1.set_text(name1)
        entry2.set_text(name2)
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("name1", "name2", "name_btn1", "name_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        entry1 = self.get_widget("type1")
        entry2 = self.get_widget("type2")
        entry1.set_text(str(self.rp1.get_type()))
        entry2.set_text(str(self.rp2.get_type()))
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("type1", "type2", "type_btn1", "type_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        gramps1 = self.rp1.get_gramps_id()
        gramps2 = self.rp2.get_gramps_id()
        entry1 = self.get_widget("gramps1")
        entry2 = self.get_widget("gramps2")
        entry1.set_text(gramps1)
        entry2.set_text(gramps2)
        if entry1.get_text() == entry2.get_text():
            for widget_name in ("gramps1", "gramps2", "gramps_btn1", "gramps_btn2"):
                self.get_widget(widget_name).set_sensitive(False)

        # Main window widgets that determine which handle survives
        rbutton1 = self.get_widget("handle_btn1")
        rbutton_label1 = self.get_widget("label_handle_btn1")
        rbutton_label2 = self.get_widget("label_handle_btn2")
        rbutton_label1.set_label("%s [%s]" % (name1, gramps1))
        rbutton_label2.set_label("%s [%s]" % (name2, gramps2))
        rbutton1.connect("toggled", self.on_handle1_toggled)

        self.connect_button("repository_help", self.cb_help)
        self.connect_button("repository_ok", self.cb_merge)
        self.connect_button("repository_cancel", self.close)

        self.show()

    def on_handle1_toggled(self, obj):
        """preferred repository changes"""
        if obj.get_active():
            self.get_widget("name_btn1").set_active(True)
            self.get_widget("type_btn1").set_active(True)
            self.get_widget("gramps_btn1").set_active(True)
        else:
            self.get_widget("name_btn2").set_active(True)
            self.get_widget("type_btn2").set_active(True)
            self.get_widget("gramps_btn2").set_active(True)

    def cb_help(self, obj):
        """Display the relevant portion of the Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def cb_merge(self, obj):
        """
        Perform the merge of the repositories when the merge button is clicked.
        """
        self.uistate.set_busy_cursor(True)
        use_handle1 = self.get_widget("handle_btn1").get_active()
        if use_handle1:
            phoenix = self.rp1
            titanic = self.rp2
        else:
            phoenix = self.rp2
            titanic = self.rp1

        if self.get_widget("name_btn1").get_active() ^ use_handle1:
            phoenix.set_name(titanic.get_name())
        if self.get_widget("type_btn1").get_active() ^ use_handle1:
            phoenix.set_type(titanic.get_type())
        if self.get_widget("gramps_btn1").get_active() ^ use_handle1:
            phoenix.set_gramps_id(titanic.get_gramps_id())

        query = MergeRepositoryQuery(self.dbstate, phoenix, titanic)
        query.execute()
        # Add the selected handle to history so that when merge is complete,
        # phoenix is the selected row.
        self.uistate.set_active(phoenix.get_handle(), "Repository")
        self.uistate.set_busy_cursor(False)
        self.close()



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Andrew I Baznikin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman, largely based on relationship.py by Don Allingham
# and on valuable input from Lars Kr. Lundin
"""
Specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------
#
# Danish-specific definitions of relationships
#
# -------------------------------------------------------------------------

_level_name = [
    "",
    "første",
    "anden",
    "tredje",
    "fjerde",
    "femte",
    "sjette",
    "syvende",
    "ottende",
    "niende",
    "tiende",
    "ellevte",
    "tolvte",
    "trettende",
    "fjortende",
    "femtende",
    "sekstende",
    "syttende",
    "attende",
    "nittende",
    "tyvende",
    "enogtyvende",
    "toogtyvende",
    "treogtyvende",
    "fireogtyvende",
    "femogtyvende",
    "seksogtyvende",
    "syvogtyvende",
    "otteogtyvende",
    "niogtyvende",
    "tredivte",
]

_parents_level = [
    "forældre",
    "bedsteforældre",
    "oldeforældre",
    "tipoldeforældre",
    "tiptipoldeforældre",
    "tiptiptipoldeforældre",
]

_father_level = [
    "",
    "faderen",
    "bedstefaderen",
    "oldefaderen",
    "tipoldefaderen",
]

_mother_level = [
    "",
    "moderen",
    "bedstemoderen",
    "oldemoderen",
    "tipoldemoderen",
]

_son_level = [
    "",
    "sønnen",
    "barnebarnet",
    "oldebarnet",
]

_daughter_level = [
    "",
    "datteren",
    "barnebarnet",
    "oldebarnet",
]

_sister_level = [
    "",
    "søsteren",
    "tanten",
    "grandtanten",
    "oldetanten",
]

_brother_level = [
    "",
    "broderen",
    "onklen",
    "grandonklen",
    "oldeonkel",
]

_nephew_level = [
    "",
    "nevøen",
    "næstsøskendebarnet",
    "broderens barnebarn",
]

_niece_level = [
    "",
    "niecen",
    "næstsøskendebarnet",
    "søsterens barnebarn",
]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_parents(self, level):
        if level > len(_parents_level) - 1:
            # return "fjern forfader"
            # Instead of "remote ancestors" using "tip (level) oldeforældre" here.
            return "tip (%d) oldeforældre" % level
        else:
            return _parents_level[level]

    def pair_up(self, rel_list):
        result = []
        item = ""
        for word in rel_list[:]:
            if not word:
                continue
            if item:
                if word == "søster":
                    item = item[0:-1]
                    word = "ster"
                elif word == "sønne":
                    word = "søn"
                result.append(item + word)
                item = ""
            else:
                item = word
        if item:
            result.append(item)
        gen_result = [item + "s" for item in result[0:-1]]
        return " ".join(gen_result + result[-1:])

    def get_direct_ancestor(self, person, rel_string):
        result = []
        for rel in rel_string:
            if rel == "f":
                result.append("far")
            else:
                result.append("mor")
        return self.pair_up(result)

    def get_direct_descendant(self, person, rel_string):
        result = []
        for ix in range(len(rel_string) - 2, -1, -1):
            if rel_string[ix] == "f":
                result.append("sønne")
            else:
                result.append("datter")
        if person == Person.MALE:
            result.append("søn")
        else:
            result.append("datter")
        return self.pair_up(result)

    def get_two_way_rel(self, person, first_rel_string, second_rel_string):
        result = []
        for ix in range(len(second_rel_string) - 1):
            if second_rel_string[ix] == "f":
                result.append("far")
            else:
                result.append("mor")
        if len(first_rel_string) > 1:
            if first_rel_string[-2] == "f":
                result.append("bror")
            else:
                result.append("søster")
            for ix in range(len(first_rel_string) - 3, -1, -1):
                if first_rel_string[ix] == "f":
                    result.append("sønne")
                else:
                    result.append("datter")
            if person == Person.MALE:
                result.append("søn")
            else:
                result.append("datter")
        else:
            if person == Person.MALE:
                result.append("bror")
            else:
                result.append("søster")
        return self.pair_up(result)

    def get_relationship(self, secondRel, firstRel, orig_person, other_person):
        common = ""
        if not firstRel:
            if not secondRel:
                return ("", common)
            else:
                return (self.get_direct_ancestor(other_person, secondRel), common)
        elif not secondRel:
            return (self.get_direct_descendant(other_person, firstRel), common)
        else:
            return (self.get_two_way_rel(other_person, firstRel, secondRel), common)

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        return self.get_relationship(reltocommon_a, reltocommon_b, gender_a, gender_b)[
            0
        ]

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        return self.get_two_way_rel(gender_b, "", "")


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_da.py
    # (Above not needed here)

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasIdOf
#
# -------------------------------------------------------------------------
class MatchIdOf(Rule):
    """Rule that checks for a person with a specific Gramps ID"""

    labels = [_("ID:")]
    name = _("Person with <Id>")
    description = _("Matches person with a specified Gramps ID")
    category = _("General filters")

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.gramps_id.find(self.list[0]) != -1



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide calendar to sdn (serial date number) conversion.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import math

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_GRG_SDN_OFFSET = 32045
_GRG_DAYS_PER_5_MONTHS = 153
_GRG_DAYS_PER_4_YEARS = 1461
_GRG_DAYS_PER_400_YEARS = 146097

_JLN_SDN_OFFSET = 32083
_JLN_DAYS_PER_5_MONTHS = 153
_JLN_DAYS_PER_4_YEARS = 1461

_HBR_HALAKIM_PER_HOUR = 1080
_HBR_HALAKIM_PER_DAY = 25920
_HBR_HALAKIM_PER_LUNAR_CYCLE = 29 * _HBR_HALAKIM_PER_DAY + 13753
_HBR_HALAKIM_PER_METONIC_CYCLE = _HBR_HALAKIM_PER_LUNAR_CYCLE * (12 * 19 + 7)
_HBR_SDN_OFFSET = 347997
_HBR_NEW_MOON_OF_CREATION = 31524
_HBR_NOON = 18 * _HBR_HALAKIM_PER_HOUR
_HBR_AM3_11_20 = (9 * _HBR_HALAKIM_PER_HOUR) + 204
_HBR_AM9_32_43 = (15 * _HBR_HALAKIM_PER_HOUR) + 589

_HBR_SUNDAY = 0
_HBR_MONDAY = 1
_HBR_TUESDAY = 2
_HBR_WEDNESDAY = 3
_HBR_FRIDAY = 5

_HBR_MONTHS_PER_YEAR = [
    12,
    12,
    13,
    12,
    12,
    13,
    12,
    13,
    12,
    12,
    13,
    12,
    12,
    13,
    12,
    12,
    13,
    12,
    13,
]

_HBR_YEAR_OFFSET = [
    0,
    12,
    24,
    37,
    49,
    61,
    74,
    86,
    99,
    111,
    123,
    136,
    148,
    160,
    173,
    185,
    197,
    210,
    222,
]

_FR_SDN_OFFSET = 2375474
_FR_DAYS_PER_4_YEARS = 1461
_FR_DAYS_PER_MONTH = 30
_PRS_EPOCH = 1948320.5
_ISM_EPOCH = 1948439.5


def _tishri1(metonic_year, molad_day, molad_halakim):
    tishri1 = molad_day
    dow = tishri1 % 7
    leap_year = metonic_year in [2, 5, 7, 10, 13, 16, 18]
    last_was_leap_year = metonic_year in [3, 6, 8, 11, 14, 17, 0]

    # Apply rules 2, 3 and 4.
    if (
        (molad_halakim >= _HBR_NOON)
        or ((not leap_year) and dow == _HBR_TUESDAY and molad_halakim >= _HBR_AM3_11_20)
        or (
            last_was_leap_year
            and dow == _HBR_MONDAY
            and molad_halakim >= _HBR_AM9_32_43
        )
    ):
        tishri1 += 1
        dow += 1
        if dow == 7:
            dow = 0

    # Apply rule 1 after the others because it can cause an additional
    # delay of one day
    if dow in [_HBR_WEDNESDAY, _HBR_FRIDAY, _HBR_SUNDAY]:
        tishri1 += 1

    return tishri1


def _tishri_molad(input_day):
    """
    Estimate the metonic cycle number.

    Note that this may be an under estimate because there are 6939.6896 days
    in a metonic cycle not 6940, but it will never be an over estimate. The
    loop below will correct for any error in this estimate.
    """

    metonic_cycle = (input_day + 310) // 6940

    # Calculate the time of the starting molad for this metonic cycle.

    (molad_day, molad_halakim) = _molad_of_metonic_cycle(metonic_cycle)

    # If the above was an under estimate, increment the cycle number until
    # the correct one is found.  For modern dates this loop is about 98.6%
    # likely to not execute, even once, because the above estimate is
    # really quite close.

    while molad_day < (input_day - 6940 + 310):
        metonic_cycle += 1
        molad_halakim += _HBR_HALAKIM_PER_METONIC_CYCLE
        molad_day += molad_halakim // _HBR_HALAKIM_PER_DAY
        molad_halakim = molad_halakim % _HBR_HALAKIM_PER_DAY

    # Find the molad of Tishri closest to this date.

    for metonic_year in range(0, 20):
        if molad_day > input_day - 74:
            break

        molad_halakim += (
            _HBR_HALAKIM_PER_LUNAR_CYCLE * _HBR_MONTHS_PER_YEAR[metonic_year]
        )
        molad_day += molad_halakim // _HBR_HALAKIM_PER_DAY
        molad_halakim = molad_halakim % _HBR_HALAKIM_PER_DAY

    return (metonic_cycle, metonic_year, molad_day, molad_halakim)


def _molad_of_metonic_cycle(metonic_cycle):
    """
    Start with the time of the first molad after creation.
    """

    r1 = _HBR_NEW_MOON_OF_CREATION

    # Calculate metonic_cycle * HALAKIM_PER_METONIC_CYCLE.  The upper 32
    # bits of the result will be in r2 and the lower 16 bits will be
    # in r1.

    r1 = r1 + (metonic_cycle * (_HBR_HALAKIM_PER_METONIC_CYCLE & 0xFFFF))
    r2 = r1 >> 16
    r2 = r2 + (metonic_cycle * ((_HBR_HALAKIM_PER_METONIC_CYCLE >> 16) & 0xFFFF))

    # Calculate r2r1 / HALAKIM_PER_DAY.  The remainder will be in r1, the
    # upper 16 bits of the quotient will be in d2 and the lower 16 bits
    # will be in d1.

    d2 = r2 // _HBR_HALAKIM_PER_DAY
    r2 -= d2 * _HBR_HALAKIM_PER_DAY
    r1 = (r2 << 16) | (r1 & 0xFFFF)
    d1 = r1 // _HBR_HALAKIM_PER_DAY
    r1 -= d1 * _HBR_HALAKIM_PER_DAY

    molad_day = (d2 << 16) | d1
    molad_halakim = r1

    return (molad_day, molad_halakim)


def _start_of_year(year):
    """
    Calculate the start of the year.
    """
    metonic_cycle = (year - 1) // 19
    metonic_year = (year - 1) % 19
    (molad_day, molad_halakim) = _molad_of_metonic_cycle(metonic_cycle)

    molad_halakim = molad_halakim + (
        _HBR_HALAKIM_PER_LUNAR_CYCLE * _HBR_YEAR_OFFSET[metonic_year]
    )
    molad_day = molad_day + (molad_halakim // _HBR_HALAKIM_PER_DAY)
    molad_halakim = molad_halakim % _HBR_HALAKIM_PER_DAY

    ptishri1 = _tishri1(metonic_year, molad_day, molad_halakim)

    return (metonic_cycle, metonic_year, molad_day, molad_halakim, ptishri1)


def hebrew_sdn(year, month, day):
    """Convert a Jewish calendar date to an SDN number."""

    if month in [1, 2]:
        # It is Tishri or Heshvan - don't need the year length.
        (
            metonic_cycle,
            metonic_year,
            molad_day,
            molad_halakim,
            tishri1,
        ) = _start_of_year(year)
        if month == 1:
            sdn = tishri1 + day - 1
        else:
            sdn = tishri1 + day + 29
    elif month == 3:
        # It is Kislev - must find the year length.

        # Find the start of the year.
        (
            metonic_cycle,
            metonic_year,
            molad_day,
            molad_halakim,
            tishri1,
        ) = _start_of_year(year)

        # Find the end of the year.
        molad_halakim = molad_halakim + (
            _HBR_HALAKIM_PER_LUNAR_CYCLE * _HBR_MONTHS_PER_YEAR[metonic_year]
        )
        molad_day = molad_day + (molad_halakim // _HBR_HALAKIM_PER_DAY)
        molad_halakim = molad_halakim % _HBR_HALAKIM_PER_DAY
        tishri1_after = _tishri1((metonic_year + 1) % 19, molad_day, molad_halakim)

        year_length = tishri1_after - tishri1

        if year_length in [355, 385]:
            sdn = tishri1 + day + 59
        else:
            sdn = tishri1 + day + 58
    elif month in [4, 5, 6]:
        # It is Tevet, Shevat or Adar I - don't need the year length

        (
            metonic_cycle,
            metonic_year,
            molad_day,
            molad_halakim,
            tishri1_after,
        ) = _start_of_year(year + 1)

        if _HBR_MONTHS_PER_YEAR[(year - 1) % 19] == 12:
            length_of_adar_1and2 = 29
        else:
            length_of_adar_1and2 = 59

        if month == 4:
            sdn = tishri1_after + day - length_of_adar_1and2 - 237
        elif month == 5:
            sdn = tishri1_after + day - length_of_adar_1and2 - 208
        else:
            sdn = tishri1_after + day - length_of_adar_1and2 - 178
    else:
        # It is Adar II or later - don't need the year length.
        (
            metonic_cycle,
            metonic_year,
            molad_day,
            molad_halakim,
            tishri1_after,
        ) = _start_of_year(year + 1)

        if month == 7:
            sdn = tishri1_after + day - 207
        elif month == 8:
            sdn = tishri1_after + day - 178
        elif month == 9:
            sdn = tishri1_after + day - 148
        elif month == 10:
            sdn = tishri1_after + day - 119
        elif month == 11:
            sdn = tishri1_after + day - 89
        elif month == 12:
            sdn = tishri1_after + day - 60
        elif month == 13:
            sdn = tishri1_after + day - 30
        else:
            return 0
    return sdn + _HBR_SDN_OFFSET


def hebrew_ymd(sdn):
    """Convert an SDN number to a Hebrew calendar date."""

    input_day = sdn - _HBR_SDN_OFFSET
    # TODO if input_day <= 0, the result is a date invalid in Hebrew calendar!

    (metonic_cycle, metonic_year, day1, halakim) = _tishri_molad(input_day)
    tishri1 = _tishri1(metonic_year, day1, halakim)

    if input_day >= tishri1:
        # It found Tishri 1 at the start of the year

        year = (metonic_cycle * 19) + metonic_year + 1
        if input_day < tishri1 + 59:
            if input_day < tishri1 + 30:
                month = 1
                day = input_day - tishri1 + 1
            else:
                month = 2
                day = input_day - tishri1 - 29
            return (year, month, day)

        # We need the length of the year to figure this out, so find
        # Tishri 1 of the next year.

        halakim += _HBR_HALAKIM_PER_LUNAR_CYCLE * _HBR_MONTHS_PER_YEAR[metonic_year]
        day1 += halakim // _HBR_HALAKIM_PER_DAY
        halakim = halakim % _HBR_HALAKIM_PER_DAY
        tishri1_after = _tishri1((metonic_year + 1) % 19, day1, halakim)
    else:
        # It found Tishri 1 at the end of the year.

        year = metonic_cycle * 19 + metonic_year
        if input_day >= tishri1 - 177:
            # It is one of the last 6 months of the year.
            if input_day > tishri1 - 30:
                month = 13
                day = input_day - tishri1 + 30
            elif input_day > tishri1 - 60:
                month = 12
                day = input_day - tishri1 + 60
            elif input_day > tishri1 - 89:
                month = 11
                day = input_day - tishri1 + 89
            elif input_day > tishri1 - 119:
                month = 10
                day = input_day - tishri1 + 119
            elif input_day > tishri1 - 148:
                month = 9
                day = input_day - tishri1 + 148
            else:
                month = 8
                day = input_day - tishri1 + 178
            return (year, month, day)

        if _HBR_MONTHS_PER_YEAR[(year - 1) % 19] == 13:
            month = 7
            day = input_day - tishri1 + 207
            if day > 0:
                return (year, month, day)
            month -= 1
            day += 30
            if day > 0:
                return (year, month, day)
            month -= 1
            day += 30
        else:
            month = 6
            day = input_day - tishri1 + 207
            if day > 0:
                return (year, month, day)
            month -= 1
            day += 30

        if day > 0:
            return (year, month, day)
        month -= 1
        day += 29
        if day > 0:
            return (year, month, day)

        # We need the length of the year to figure this out, so find
        # Tishri 1 of this year
        tishri1_after = tishri1
        (metonic_cycle, metonic_year, day1, halakim) = _tishri_molad(day1 - 365)
        tishri1 = _tishri1(metonic_year, day1, halakim)

    year_length = tishri1_after - tishri1
    day = input_day - tishri1 - 29
    if year_length in [355, 385]:
        # Heshvan has 30 days
        if day <= 30:
            month = 2
            return (year, month, day)
        day -= 30
    else:
        # Heshvan has 29 days
        if day <= 29:
            month = 2
            return (year, month, day)

        day -= 29

    # It has to be Kislev
    return (year, 3, day)


def julian_sdn(year, month, day):
    """Convert a Julian calendar date to an SDN number."""

    if year < 0:
        year += 4801
    else:
        year += 4800

    # Adjust the start of the year
    if month > 2:
        month -= 3
    else:
        month += 9
        year -= 1

    return (
        (year * _JLN_DAYS_PER_4_YEARS) // 4
        + (month * _JLN_DAYS_PER_5_MONTHS + 2) // 5
        + day
        - _JLN_SDN_OFFSET
    )


def julian_ymd(sdn):
    """Convert an SDN number to a Julian date."""
    temp = (sdn + _JLN_SDN_OFFSET) * 4 - 1

    # Calculate the year and day of year (1 <= day_of_year <= 366)
    year = temp // _JLN_DAYS_PER_4_YEARS
    day_of_year = (temp % _JLN_DAYS_PER_4_YEARS) // 4 + 1

    # Calculate the month and day of month
    temp = day_of_year * 5 - 3
    month = temp // _JLN_DAYS_PER_5_MONTHS
    day = (temp % _JLN_DAYS_PER_5_MONTHS) // 5 + 1

    # Convert to the normal beginning of the year
    if month < 10:
        month += 3
    else:
        year += 1
        month -= 9

    # Adjust to the B.C./A.D. type numbering
    year -= 4800
    if year <= 0:
        year -= 1

    return (year, month, day)


def gregorian_sdn(year, month, day):
    """Convert a gregorian date to an SDN number."""
    if year < 0:
        year += 4801
    else:
        year += 4800

    # Adjust the start of the year
    if month > 2:
        month -= 3
    else:
        month += 9
        year -= 1

    return (
        ((year // 100) * _GRG_DAYS_PER_400_YEARS) // 4
        + ((year % 100) * _GRG_DAYS_PER_4_YEARS) // 4
        + (month * _GRG_DAYS_PER_5_MONTHS + 2) // 5
        + day
        - _GRG_SDN_OFFSET
    )


def gregorian_ymd(sdn):
    """Convert an SDN number to a gregorian date."""
    temp = (_GRG_SDN_OFFSET + sdn) * 4 - 1

    # Calculate the century (year/100)
    century = temp // _GRG_DAYS_PER_400_YEARS

    # Calculate the year and day of year (1 <= day_of_year <= 366)
    temp = ((temp % _GRG_DAYS_PER_400_YEARS) // 4) * 4 + 3
    year = (century * 100) + (temp // _GRG_DAYS_PER_4_YEARS)
    day_of_year = (temp % _GRG_DAYS_PER_4_YEARS) // 4 + 1

    # Calculate the month and day of month
    temp = day_of_year * 5 - 3
    month = temp // _GRG_DAYS_PER_5_MONTHS
    day = (temp % _GRG_DAYS_PER_5_MONTHS) // 5 + 1

    # Convert to the normal beginning of the year
    if month < 10:
        month = month + 3
    else:
        year = year + 1
        month = month - 9

    # Adjust to the B.C./A.D. type numbering
    year = year - 4800
    if year <= 0:
        year = year - 1
    return (year, month, day)


def _check_republican_period(sdn, restrict_period):
    # French Republican calendar wasn't in use before 22.9.1792 or
    # after 1.1.1806
    if restrict_period and (sdn < 2375840 or sdn > 2380688):
        raise ValueError("Outside of the French Republican period")


def french_sdn(year, month, day, restrict_period=False):
    """Convert a French Republican Calendar date to an SDN number."""
    sdn = (
        (year * _FR_DAYS_PER_4_YEARS) // 4
        + (month - 1) * _FR_DAYS_PER_MONTH
        + day
        + _FR_SDN_OFFSET
    )
    _check_republican_period(sdn, restrict_period)
    return sdn


def french_ymd(sdn, restrict_period=False):
    """Convert an SDN number to a French Republican Calendar date."""
    _check_republican_period(sdn, restrict_period)
    temp = (sdn - _FR_SDN_OFFSET) * 4 - 1
    year = temp // _FR_DAYS_PER_4_YEARS
    day_of_year = (temp % _FR_DAYS_PER_4_YEARS) // 4
    month = (day_of_year // _FR_DAYS_PER_MONTH) + 1
    day = (day_of_year % _FR_DAYS_PER_MONTH) + 1
    return (year, month, day)


def persian_sdn(year, month, day):
    """Convert a Persian date to an SDN number."""
    if year >= 0:
        epbase = year - 474
    else:
        epbase = year - 473

    epyear = 474 + epbase % 2820

    if month <= 7:
        v1 = (month - 1) * 31
    else:
        v1 = ((month - 1) * 30) + 6
    v2 = ((epyear * 682) - 110) // 2816
    v3 = (epyear - 1) * 365 + day
    v4 = (epbase // 2820) * 1029983

    return int(math.ceil(v1 + v2 + v3 + v4 + _PRS_EPOCH - 1))


def persian_ymd(sdn):
    """Convert an SDN number to a Persian calendar date."""
    # The following is commented out and is related to bug 12576
    # sdn = math.floor(sdn) + 0.5         # float

    depoch = sdn - 2121446  # float
    cycle = math.floor(depoch / 1029983)  # int
    cyear = depoch % 1029983  # int
    if cyear == 1029982:
        ycycle = 2820
    else:
        aux1 = cyear // 366  # int
        aux2 = cyear % 366  # int
        ycycle = (((2134 * aux1) + (2816 * aux2) + 2815) // 1028522) + aux1 + 1

    year = ycycle + (2820 * cycle) + 474
    if year <= 0:
        year = year - 1

    yday = sdn - persian_sdn(year, 1, 1) + 1  # float !
    if yday < 186:
        month = math.ceil(yday / 31)
    else:
        month = math.ceil((yday - 6) / 30)
    day = (sdn - persian_sdn(year, month, 1)) + 1
    return (int(year), int(month), int(day))


def islamic_sdn(year, month, day):
    """Convert an Islamic date to an SDN number."""
    v1 = math.ceil(29.5 * (month - 1))
    v2 = (year - 1) * 354
    v3 = math.floor((3 + (11 * year)) // 30)

    return int(math.ceil((day + v1 + v2 + v3 + _ISM_EPOCH) - 1))


def islamic_ymd(sdn):
    """Convert an SDN number to an Islamic calendar date."""
    sdn = math.floor(sdn) + 0.5
    year = int(math.floor(((30 * (sdn - _ISM_EPOCH)) + 10646) / 10631))
    month = int(min(12, math.ceil((sdn - (29 + islamic_sdn(year, 1, 1))) / 29.5) + 1))
    day = int((sdn - islamic_sdn(year, month, 1)) + 1)
    return (year, month, day)


def swedish_sdn(year, month, day):
    """Convert a Swedish date to an SDN number."""
    datum = (year, month, day)
    # Swedish Calendar
    if (1700, 3, 1) <= datum <= (1712, 2, 30):
        return julian_sdn(year, month, day) - 1
    # Gregorian Calendar (1753-03-01)
    if (1753, 3, 1) <= datum:
        return gregorian_sdn(year, month, day)
    return julian_sdn(year, month, day)


def swedish_ymd(sdn):
    """Convert an SDN number to a Swedish calendar date."""
    if sdn == 2346425:
        return (1712, 2, 30)
    # Swedish Calendar
    if 2342042 <= sdn < 2346425:
        return julian_ymd(sdn + 1)
    # Gregorian Calendar (1753-03-01)
    if sdn >= 2361390:
        return gregorian_ymd(sdn)
    return julian_ymd(sdn)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009   Stephane Charette
# Copyright (C) 2019-       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Find possible leading and/or trailing spaces in places name and people"

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.utils.place import conv_lat_lon
from gramps.gui.plug import tool
from gramps.gui.editors import EditPlace, EditPerson
from gramps.gen.errors import WindowActiveError
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.utils import ProgressMeter
from gramps.gui.display import display_help
from gramps.gui.glade import Glade
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Remove_leading_and_trailing_spaces", "manual")


def validate_lat_lon(field):
    """
    Return True if some characters are found in the field
    # hyphen (u+2010)
    # non-breaking hyphen (u+2011)
    # figure dash (u+2012)
    # en dash (u+2013)
    # em dash (u+2014)
    # horizontal bar (u+2015)
    """
    for char in (",", "\u2010", "\u2011", "\u2012", "\u2013", "\u2014", "\u2015"):
        if field.find(char) != -1:
            return True
    return False


# ------------------------------------------------------------------------
#
# RemoveSpaces class
#
# ------------------------------------------------------------------------
class RemoveSpaces(ManagedWindow):
    """
    Find leading and trailing spaces in Place names and person names
    """

    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate
        dummy_opt = options_class
        dummy_nme = name
        dummy_cb = callback

        self.title = _("Clean input data")
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.dbstate = dbstate
        self.uistate = uistate
        self.db = dbstate.db
        self.tooltip = ""
        self.f_lat = False
        self.f_lon = False

        top_dialog = Glade()

        top_dialog.connect_signals(
            {
                "destroy_passed_object": self.close,
                "on_help_clicked": self.on_help_clicked,
                "on_delete_event": self.close,
            }
        )

        window = top_dialog.toplevel
        title = top_dialog.get_object("title")
        self.set_window(window, title, self.title)
        tip = _(
            "Search leading and/or trailing spaces for persons"
            " and places. Search comma or bad sign in coordinates"
            " fields.\n"
            "Double click on a row to edit its content."
        )
        title.set_tooltip_text(tip)

        # start the progress indicator
        self.progress = ProgressMeter(self.title, _("Starting"), parent=uistate.window)
        steps = self.db.get_number_of_people() + self.db.get_number_of_places()
        self.progress.set_pass(
            _("Looking for possible fields with leading or" " trailing spaces"), steps
        )

        self.model_1 = Gtk.ListStore(
            GObject.TYPE_STRING,  # 0==handle
            GObject.TYPE_STRING,  # 1==firstname
            GObject.TYPE_STRING,  # 2==surname
            GObject.TYPE_STRING,  # 3==alternate name
            GObject.TYPE_STRING,  # 4==group_as
        )
        self.model_1.set_sort_column_id(Gtk.TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, 1)

        self.label_1 = top_dialog.get_object("label_1")
        self.label_1.set_text(_("Person"))
        self.treeview_1 = top_dialog.get_object("treeview_1")
        self.treeview_1.set_model(self.model_1)
        col1 = Gtk.TreeViewColumn(_("handle"), Gtk.CellRendererText(), text=0)
        renderer1 = Gtk.CellRendererText()
        renderer1.set_property("underline-set", True)
        renderer1.set_property("underline", 2)  # 2=double underline
        col2 = Gtk.TreeViewColumn(_("firstname"), renderer1, text=1)
        renderer2 = Gtk.CellRendererText()
        renderer2.set_property("underline-set", True)
        renderer2.set_property("underline", 2)  # 2=double underline
        col3 = Gtk.TreeViewColumn(_("surname"), renderer2, text=2)
        renderer3 = Gtk.CellRendererText()
        renderer3.set_property("underline-set", True)
        renderer3.set_property("underline", 2)  # 2=double underline
        col4 = Gtk.TreeViewColumn(_("alternate name"), renderer3, text=3)
        renderer4 = Gtk.CellRendererText()
        renderer4.set_property("underline-set", True)
        renderer4.set_property("underline", 2)  # 2=double underline
        col5 = Gtk.TreeViewColumn(_("group as"), renderer4, text=4)
        col1.set_resizable(True)
        col1.set_visible(False)
        col2.set_resizable(True)
        col3.set_resizable(True)
        col4.set_resizable(True)
        col5.set_resizable(True)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col3.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col4.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col5.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.treeview_1.append_column(col1)
        self.treeview_1.append_column(col2)
        self.treeview_1.append_column(col3)
        self.treeview_1.append_column(col4)
        self.treeview_1.append_column(col5)
        self.treeselection = self.treeview_1.get_selection()
        self.treeview_1.connect("row-activated", self.rowactivated_cb1)

        self.model_2 = Gtk.ListStore(
            GObject.TYPE_STRING,  # 0==handle
            GObject.TYPE_STRING,  # 1==name
            GObject.TYPE_STRING,  # 2==latitude
            GObject.TYPE_STRING,  # 3==longitude
            GObject.TYPE_STRING,
        )  # 4==tooltip
        self.model_2.set_sort_column_id(Gtk.TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, 1)

        self.label_2 = top_dialog.get_object("label_2")
        self.label_2.set_text(_("Place"))
        self.treeview_2 = top_dialog.get_object("treeview_2")
        self.treeview_2.set_model(self.model_2)
        col1 = Gtk.TreeViewColumn(_("handle"), Gtk.CellRendererText(), text=0)
        renderer5 = Gtk.CellRendererText()
        renderer5.set_property("underline-set", True)
        renderer5.set_property("underline", 2)  # 2=double underline
        col2 = Gtk.TreeViewColumn(_("name"), renderer5, text=1)
        renderer6 = Gtk.CellRendererText()
        renderer6.set_property("underline-set", True)
        renderer6.set_property("underline", 2)  # 2=double underline
        col3 = Gtk.TreeViewColumn(_("latitude"), renderer6, text=2)
        renderer7 = Gtk.CellRendererText()
        renderer7.set_property("underline-set", True)
        renderer7.set_property("underline", 2)  # 2=double underline
        col4 = Gtk.TreeViewColumn(_("longitude"), renderer7, text=3)
        col5 = Gtk.TreeViewColumn(_("tooltip"), Gtk.CellRendererText(), text=4)
        col1.set_resizable(True)
        col1.set_visible(False)
        col2.set_resizable(True)
        col3.set_resizable(True)
        col4.set_resizable(True)
        col5.set_visible(False)
        col1.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col2.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col3.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col4.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col5.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.treeview_2.append_column(col1)
        self.treeview_2.append_column(col2)
        self.treeview_2.append_column(col3)
        self.treeview_2.append_column(col4)
        self.treeview_2.append_column(col5)
        self.treeview_2.set_tooltip_column(4)
        self.treeselection = self.treeview_2.get_selection()
        self.treeview_2.connect("row-activated", self.rowactivated_cb2)

        self.places()
        self.people()

        # close the progress bar
        self.progress.close()

        self.show()

    def add_text_to_tip(self, nll, mess):
        """
        Add a text to tooltip for places.
        nll: 0: name
             1: latitude
             2: longitude
        mess: text to add
        """
        if nll == 0:
            self.tooltip = _("Name:", "place")
            self.tooltip += " "
            self.tooltip += mess
        elif nll == 1:
            if not self.f_lat:
                if self.tooltip != "":
                    self.tooltip += "\n"
                self.tooltip += _("Latitude:")
                self.tooltip += " "
                self.f_lat = True
            else:
                self.tooltip += ", "
            self.tooltip += mess
        elif nll == 2:
            if not self.f_lon:
                if self.tooltip != "":
                    self.tooltip += "\n"
                self.tooltip += _("Longitude:")
                self.tooltip += " "
                self.f_lon = True
            else:
                self.tooltip += ", "
            self.tooltip += mess

    def places(self):
        """
        For all places in the database, if the name contains leading or
        trailing spaces.
        """
        for place_handle in self.db.get_place_handles():
            self.progress.step()
            place = self.db.get_place_from_handle(place_handle)
            place_name = place.get_name()
            pname = place_name.get_value()
            found = False
            self.tooltip = ""
            if pname != pname.strip():
                self.add_text_to_tip(0, _("leading and/or trailing spaces"))
                found = True
            plat = place.get_latitude()
            plon = place.get_longitude()
            lat, lon = conv_lat_lon(plat, plon, "D.D8")
            self.f_lat = self.f_lon = False
            if not lat:
                self.add_text_to_tip(1, _("invalid format"))
                found = True
            if plat != plat.strip():
                self.add_text_to_tip(1, _("leading and/or trailing spaces"))
                found = True
            if plat.find(",") != -1:
                self.add_text_to_tip(1, _("comma instead of dot"))
                found = True
            if validate_lat_lon(plat):
                self.add_text_to_tip(1, _("invalid char instead of '-'"))
                found = True
            if not lon:
                self.add_text_to_tip(2, _("invalid format"))
                found = True
            if plon != plon.strip():
                self.add_text_to_tip(2, _("leading and/or trailing spaces"))
                found = True
            if plon.find(",") != -1:
                self.add_text_to_tip(2, _("comma instead of dot"))
                found = True
            if validate_lat_lon(plat):
                self.add_text_to_tip(1, _("invalid char instead of '-'"))
                found = True
            if found:
                value = (place_handle, pname, plat, plon, self.tooltip)
                self.model_2.append(value)
        return True

    def people(self):
        """
        For all persons in the database, if the name contains leading or
        trailing spaces. Works for alternate names and group_as.
        """
        for person_handle in self.db.get_person_handles():
            self.progress.step()
            person = self.db.get_person_from_handle(person_handle)
            primary_name = person.get_primary_name()
            fname = primary_name.get_first_name()
            found = False
            if fname != fname.strip():
                found = True
            sname = primary_name.get_primary_surname().get_surname()
            if sname != sname.strip():
                found = True
            paname = ""
            for name in primary_name.get_surname_list():
                aname = name.get_surname()
                if aname != sname and aname != aname.strip():
                    found = True
                    if paname != "":
                        paname += ", "
                    paname += aname
            groupas = primary_name.group_as
            if groupas != groupas.strip():
                found = True
            if found:
                value = (person_handle, fname, sname, paname, groupas)
                self.model_1.append(value)
        return True

    def rowactivated_cb1(self, treeview, path, column):
        """
        Called when a Person row is activated.
        """
        dummy_tv = treeview
        dummy_col = column
        iter_ = self.model_1.get_iter(path)
        handle = self.model_1.get_value(iter_, 0)
        person = self.dbstate.db.get_person_from_handle(handle)
        if person:
            try:
                EditPerson(self.dbstate, self.uistate, [], person)
            except WindowActiveError:
                pass
            return True
        return False

    def rowactivated_cb2(self, treeview, path, column):
        """
        Called when a Place row is activated.
        """
        dummy_tv = treeview
        dummy_col = column
        iter_ = self.model_2.get_iter(path)
        handle = self.model_2.get_value(iter_, 0)
        place = self.dbstate.db.get_place_from_handle(handle)
        if place:
            try:
                EditPlace(self.dbstate, self.uistate, [], place)
            except WindowActiveError:
                pass
            return True
        return False

    def on_help_clicked(self, _obj):
        """
        Display the relevant portion of Gramps manual.
        """
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def close(self, *obj):
        ManagedWindow.close(self, *obj)


# ------------------------------------------------------------------------
#
# RemoveSpacesOptions
#
# ------------------------------------------------------------------------
class RemoveSpacesOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        """Initialize the options class"""
        tool.ToolOptions.__init__(self, name, person_id)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.errors import WindowActiveError
from gramps.gen.lib import Citation, Source
from ...dbguielement import DbGUIElement
from ...selectors import SelectorFactory
from .citationrefmodel import CitationRefModel
from .embeddedlist import EmbeddedList, TEXT_COL, MARKUP_COL, ICON_COL
from ...ddtargets import DdTargets

LOG = logging.getLogger(".citation")
_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# CitationEmbedList
#
# -------------------------------------------------------------------------
class CitationEmbedList(EmbeddedList, DbGUIElement):
    """
    Citation List display tab for edit dialogs.

    Derives from the EmbeddedList class.
    """

    _HANDLE_COL = 10  # Column number from CitationRefModel
    _DND_TYPE = DdTargets.CITATION_LINK
    _DND_EXTRA = DdTargets.SOURCE_LINK

    _MSG = {
        "add": _("Create and add a new citation and new source"),
        "del": _("Remove the existing citation"),
        "edit": _("Edit the selected citation"),
        "share": _("Add an existing citation or source"),
        "up": _("Move the selected citation upwards"),
        "down": _("Move the selected citation downwards"),
    }

    # index = column in model. Value =
    #  (name, sortcol in model, width, markup/text, weigth_col
    _column_names = [
        (_("Title"), 0, 350, TEXT_COL, -1, None),
        (_("Author"), 1, 200, TEXT_COL, -1, None),
        (_("Date"), 8, 180, MARKUP_COL, -1, None),
        (_("Publisher"), 3, 200, TEXT_COL, -1, None),
        (_("Confidence"), 9, 120, MARKUP_COL, -1, None),
        (_("Page"), 5, 100, TEXT_COL, -1, None),
        (_("ID"), 6, 80, TEXT_COL, -1, None),
        (_("Private"), 7, 30, ICON_COL, -1, "gramps-lock"),
        (_("Sorted date"), 8, 80, TEXT_COL, -1, None),
        (_("Sorted confidence"), 9, 120, TEXT_COL, -1, None),
    ]

    def __init__(self, dbstate, uistate, track, data, config_key, callertitle=None):
        self.data = data
        self.callertitle = callertitle
        EmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("_Source Citations"),
            CitationRefModel,
            config_key,
            share_button=True,
            move_buttons=True,
        )
        DbGUIElement.__init__(self, dbstate.db)
        self.callman.register_handles({"citation": self.data})

    def _connect_db_signals(self):
        """
        Implement base class DbGUIElement method
        """
        # citation: citation-rebuild closes the editors, so no need to connect
        # to it
        self.callman.register_callbacks(
            {
                "citation-delete": self.citation_delete,
                "citation-update": self.citation_update,
            }
        )
        self.callman.connect_all(keys=["citation"])

    def get_icon_name(self):
        """
        Return the icon name associated with the display tab
        """
        return "gramps-source"

    def get_data(self):
        """
        Return the data associated with display tab
        """
        return self.data

    def column_order(self):
        """
        Return the column order of the columns in the display tab.
        """
        return ((1, 7), (1, 0), (1, 1), (1, 5), (1, 2), (1, 3), (1, 6), (1, 4))

    def add_button_clicked(self, obj):
        """
        Create a new Citation instance and call the EditCitation editor with
        the new citation.

        Called when the Add button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.
        """
        try:
            from .. import EditCitation

            EditCitation(
                self.dbstate,
                self.uistate,
                self.track,
                Citation(),
                Source(),
                self.add_callback,
                self.callertitle,
            )
        except WindowActiveError:
            pass

    def add_callback(self, value):
        """
        Called to update the screen when a new citation is added
        """
        data = self.get_data()
        data.append(value)
        self.callman.register_handles({"citation": [value]})
        self.changed = True
        self.rebuild()
        GLib.idle_add(self.tree.scroll_to_cell, len(data) - 1)

    def share_button_clicked(self, obj):
        SelectCitation = SelectorFactory("Citation")

        sel = SelectCitation(self.dbstate, self.uistate, self.track)
        objct = sel.run()
        LOG.debug("selected object: %s" % objct)
        # the object returned should either be a Source or a Citation
        if objct:
            if isinstance(objct, Source):
                try:
                    from .. import EditCitation

                    EditCitation(
                        self.dbstate,
                        self.uistate,
                        self.track,
                        Citation(),
                        objct,
                        callback=self.add_callback,
                        callertitle=self.callertitle,
                    )
                except WindowActiveError:
                    from ...dialog import WarningDialog

                    WarningDialog(
                        _("Cannot share this reference"),
                        self.__blocked_text(),
                        parent=self.uistate.window,
                    )
            elif isinstance(objct, Citation):
                try:
                    from .. import EditCitation

                    EditCitation(
                        self.dbstate,
                        self.uistate,
                        self.track,
                        objct,
                        callback=self.add_callback,
                        callertitle=self.callertitle,
                    )
                except WindowActiveError:
                    from ...dialog import WarningDialog

                    WarningDialog(
                        _("Cannot share this reference"),
                        self.__blocked_text(),
                        parent=self.uistate.window,
                    )
            else:
                raise ValueError("selection must be either source or citation")

    def __blocked_text(self):
        """
        Return the common text used when citation cannot be edited
        """
        return _(
            "This citation cannot be created at this time. "
            "Either the associated Source object is already being "
            "edited, or another citation associated with the same "
            "source is being edited.\n\nTo edit this "
            "citation, you need to close the object."
        )

    def edit_button_clicked(self, obj):
        """
        Get the selected Citation instance and call the EditCitation editor
        with the citation.

        Called when the Edit button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.
        """
        handle = self.get_selected()
        if handle:
            citation = self.dbstate.db.get_citation_from_handle(handle)
            try:
                from .. import EditCitation

                EditCitation(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    citation,
                    callertitle=self.callertitle,
                )
            except WindowActiveError:
                pass

    def citation_delete(self, del_citation_handle_list):
        """
        Outside of this tab citation objects have been deleted. Check if tab
        and object must be changed.
        Note: delete of object will cause reference on database to be removed,
            so this method need not do this
        """
        rebuild = False
        for handle in del_citation_handle_list:
            while self.data.count(handle) > 0:
                self.data.remove(handle)
                rebuild = True
        if rebuild:
            self.rebuild()

    def citation_update(self, upd_citation_handle_list):
        """
        Outside of this tab citation objects have been updated. Check if tab
        and object must be updated.
        """
        for handle in upd_citation_handle_list:
            if handle in self.data:
                self.rebuild()
                break

    def _handle_drag(self, row, handle):
        """
        A CITATION_LINK has been dragged
        """
        if handle:
            objct = self.dbstate.db.get_citation_from_handle(handle)
            if isinstance(objct, Citation):
                try:
                    from .. import EditCitation

                    EditCitation(
                        self.dbstate,
                        self.uistate,
                        self.track,
                        objct,
                        callback=self.add_callback,
                        callertitle=self.callertitle,
                    )
                except WindowActiveError:
                    from ...dialog import WarningDialog

                    WarningDialog(
                        _("Cannot share this reference"),
                        self.__blocked_text(),
                        parent=self.uistate.window,
                    )
            else:
                raise ValueError("selection must be either source or citation")

    def handle_extra_type(self, objtype, handle):
        """
        A SOURCE_LINK object has been dragged
        """
        if handle:
            objct = self.dbstate.db.get_source_from_handle(handle)
            if isinstance(objct, Source):
                try:
                    from .. import EditCitation

                    EditCitation(
                        self.dbstate,
                        self.uistate,
                        self.track,
                        Citation(),
                        objct,
                        callback=self.add_callback,
                        callertitle=self.callertitle,
                    )
                except WindowActiveError:
                    from ...dialog import WarningDialog

                    WarningDialog(
                        _("Cannot share this reference"),
                        self.__blocked_text(),
                        parent=self.uistate.window,
                    )
            else:
                raise ValueError("selection must be either source or citation")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Helper class for importing places.
"""
from collections import OrderedDict

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Place, PlaceName, PlaceType, PlaceRef


# -------------------------------------------------------------------------
#
# PlaceImport class
#
# -------------------------------------------------------------------------
class PlaceImport:
    """
    Helper class for importing places.
    """

    def __init__(self, db):
        self.db = db
        self.loc2handle = {}
        self.handle2loc = OrderedDict()

    def store_location(self, location, handle):
        """
        Store the location of a place already in the database.
        """
        self.loc2handle[location] = handle
        self.handle2loc[handle] = location

    def remove_location(self, handle):
        """
        Remove the location of a place already in the database.
        """
        if handle in self.handle2loc:
            loc = self.handle2loc[handle]
            del self.loc2handle[loc]
            del self.handle2loc[handle]

    def generate_hierarchy(self, trans):
        """
        Generate missing places in the place hierarchy.
        """
        for handle, location in self.handle2loc.items():
            # find title and type
            for type_num, name in enumerate(location):
                if name:
                    break

            loc = list(location)
            loc[type_num] = ""

            # find top parent
            parent = None
            for n in range(7):
                if loc[n]:
                    tup = tuple([""] * n + loc[n:])
                    parent = self.loc2handle.get(tup)
                    if parent:
                        break

            # create missing parent places
            if parent:
                n -= 1
            while n > type_num:
                if loc[n]:
                    # TODO for Arabic, should the next comma be translated?
                    title = ", ".join([item for item in loc[n:] if item])
                    parent = self.__add_place(loc[n], n, parent, title, trans)
                    self.loc2handle[tuple([""] * n + loc[n:])] = parent
                n -= 1

            # link to existing place
            if parent:
                place = self.db.get_place_from_handle(handle)
                placeref = PlaceRef()
                placeref.ref = parent
                place.set_placeref_list([placeref])
                self.db.commit_place(place, trans, place.get_change_time())

    def __add_place(self, name, type_num, parent, title, trans):
        """
        Add a missing place to the database.
        """
        place = Place()
        place_name = PlaceName()
        place_name.set_value(name)
        place.name = place_name
        place.title = title
        place.place_type = PlaceType(7 - type_num)
        if parent is not None:
            placeref = PlaceRef()
            placeref.ref = parent
            place.set_placeref_list([placeref])
        handle = self.db.add_place(place, trans)
        self.db.commit_place(place, trans)
        return handle



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the different event roles.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .grampstype import GrampsType

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# EventRoleType
#
# -------------------------------------------------------------------------
class EventRoleType(GrampsType):
    """
    Class representing role a participant played in an event.
    """

    UNKNOWN = -1
    CUSTOM = 0
    PRIMARY = 1
    CLERGY = 2
    CELEBRANT = 3
    AIDE = 4
    BRIDE = 5
    GROOM = 6
    WITNESS = 7
    FAMILY = 8
    INFORMANT = 9
    GODPARENT = 10
    FATHER = 11
    MOTHER = 12
    PARENT = 13
    CHILD = 14
    MULTIPLE = 15
    FRIEND = 16
    NEIGHBOR = 17
    OFFICIATOR = 18

    _CUSTOM = CUSTOM
    _DEFAULT = PRIMARY

    _DATAMAP = [
        (UNKNOWN, _("Unknown"), "Unknown"),
        (CUSTOM, _("Custom"), "Custom"),
        (PRIMARY, _("Primary", "Role"), "Primary"),
        (CLERGY, _("Clergy"), "Clergy"),
        (CELEBRANT, _("Celebrant"), "Celebrant"),
        (AIDE, _("Aide"), "Aide"),
        (BRIDE, _("Bride"), "Bride"),
        (GROOM, _("Groom"), "Groom"),
        (WITNESS, _("Witness"), "Witness"),
        (FAMILY, _("Family", "Role"), "Family"),
        (INFORMANT, _("Informant"), "Informant"),
        (GODPARENT, _("Godparent"), "Godparent"),
        (FATHER, _("Father"), "Father"),
        (MOTHER, _("Mother"), "Mother"),
        (PARENT, _("Parent"), "Parent"),
        (CHILD, _("Child"), "Child"),
        (MULTIPLE, _("Multiple birth"), "Multiple birth"),
        (FRIEND, _("Friend"), "Friend"),
        (NEIGHBOR, _("Neighbor"), "Neighbor"),
        (OFFICIATOR, _("Officiator"), "Officiator"),
    ]

    def __init__(self, value=None):
        GrampsType.__init__(self, value)

    def is_primary(self):
        """
        Returns True if EventRoleType is PRIMARY, False otherwise.
        """
        return self.value == self.PRIMARY

    def is_family(self):
        """
        Returns True if EventRoleType is FAMILY, False otherwise.
        """
        return self.value == self.FAMILY
